"""
Posterior over states after applying (Extended/Unscented) Kalman filtering/smoothing

Contains the discrete time and function outputs.
Provides dense output by being callable.
Can function values can also be accessed by indexing.
"""
from warnings import warn

import numpy as np

from probnum._randomvariablelist import _RandomVariableList
from probnum.filtsmooth.filtsmoothposterior import FiltSmoothPosterior


class KalmanPosterior(FiltSmoothPosterior):
    """
    Posterior Distribution after (Extended/Unscented) Kalman Filtering/Smoothing


    Parameters
    ----------
    locations : `array_like`
        Locations / Times of the discrete-time estimates.
    state_rvs : :obj:`list` of :obj:`RandomVariable`
        Estimated states (in the state-space model view) of the discrete-time estimates.
    gauss_filter : :obj:`GaussFiltSmooth`
        Filter/smoother used to compute the discrete-time estimates.
    """

    def __init__(self, locations, state_rvs, gauss_filter):
        self._locations = np.asarray(locations)
        self.gauss_filter = gauss_filter
        self._state_rvs = _RandomVariableList(state_rvs)

    @property
    def locations(self):
        """:obj:`np.ndarray`: Locations / times of the discrete observations"""
        return self._locations

    @property
    def state_rvs(self):
        """
        :obj:`list` of :obj:`RandomVariable`: Discrete-time posterior state estimates
        """
        return self._state_rvs

    def __call__(self, t, smoothed=True):
        """
        Evaluate the time-continuous posterior at location `t`

        Algorithm:
        1. Find closest t_prev and t_next, with t_prev < t < t_next
        2. Predict from t_prev to t
        3. (if `smoothed=True`) Predict from t to t_next
        4. (if `smoothed=True`) Smooth from t_next to t
        5. Return random variable for time t

        Parameters
        ----------
        t : float
            Location, or time, at which to evaluate the posterior.
        smoothed : bool, optional
            If ``True`` (default) perform smooth interpolation. If ``False`` perform a
            prediction from the previous location, without smoothing.

        Returns
        -------
        :obj:`RandomVariable`
            Estimate of the states at time ``t``.
        """
        if t < self.locations[0]:
            raise ValueError(
                "Invalid location; Can not compute posterior for a location earlier "
                "than the initial location"
            )

        if t in self.locations:
            idx = (self.locations <= t).sum() - 1
            discrete_estimate = self.state_rvs[idx]
            return discrete_estimate

        if self.locations[0] < t < self.locations[-1]:
            pred_rv = self._predict_to_loc(t)
            if smoothed:
                smoothed_rv = self._smooth_prediction(pred_rv, t)
                return smoothed_rv
            else:
                return pred_rv

        # else: t > self.locations[-1]:
        if smoothed:
            warn("`smoothed=True` is ignored for extrapolation.")
        return self._predict_to_loc(t)

    def _predict_to_loc(self, loc):
        """Predict states at location `loc` from the closest, previous state"""
        prev_idx = (self.locations < loc).sum() - 1
        prev_loc = self.locations[prev_idx]
        prev_rv = self.state_rvs[prev_idx]

        pred_rv, _ = self.gauss_filter.predict(
            start=prev_loc, stop=loc, randvar=prev_rv
        )
        return pred_rv

    def _smooth_prediction(self, pred_rv, loc):
        """Smooth the predicted state at location `loc` using the next closest"""
        next_idx = (self.locations < loc).sum()
        next_loc = self.locations[next_idx]
        next_rv = self._state_rvs[next_idx]
        next_pred, crosscov = self.gauss_filter.predict(
            start=loc, stop=next_loc, randvar=pred_rv
        )
        smoothed_rv = self.gauss_filter.smooth_step(
            pred_rv, next_pred, next_rv, crosscov
        )
        return smoothed_rv

    def __len__(self):
        return len(self.locations)

    def __getitem__(self, idx):
        return self.state_rvs[idx]



from typing import Union
import copy

import numpy as np


def derive_random_seed(*args: Union[np.random.RandomState, np.random.Generator]) -> int:
    def _sample(rng: Union[np.random.RandomState, np.random.Generator]) -> int:
        if isinstance(rng, np.random.RandomState):
            return copy.copy(rng).randint(0, 2 ** 32, size=None, dtype=int)
        elif isinstance(rng, np.random.Generator):
            return copy.copy(rng).integers(
                0, 2 ** 32, size=None, dtype=int, endpoint=False
            )
        else:
            raise ValueError("Unsupported type of random number generator")

    seed = _sample(args[0])

    for i in range(1, len(args)):
        seed = seed ^ _sample(args[i])

    return seed



"""
Quadrature rules based on polynomial functions.

Class of quadrature rules derived by constructing polynomial functions which are simple
to integrate.
"""

from probnum.quad.quadrature import Quadrature
from probnum import utils


class PolynomialQuadrature(Quadrature):
    """
    Quadrature rule based on polynomial functions.

    A polynomial quadrature rule is given by a collection of nodes, the roots of the
    polynomial function and a set of corresponding weights.

    Parameters
    ----------
    nodes : ndarray, shape=(n,d)
        Integration nodes.
    weights : ndarray, shape=(n,)
        Integration weights.
    bounds : ndarray, shape=(d, 2)
        Integration bounds.

    See Also
    --------
    Clenshaw-Curtis : Clenshaw-Curtis quadrature rule.
    """

    def __init__(self, nodes, weights, bounds):
        """
        Create an instance of a quadrature method.
        """
        utils.assert_is_2d_ndarray(nodes)
        utils.assert_is_1d_ndarray(weights)
        utils.assert_is_2d_ndarray(bounds)
        if len(nodes) != len(weights) or len(nodes.T) != len(bounds):
            raise ValueError(
                "Either nodes and weights or nodes and bounds are incompatible."
            )
        self.nodes = nodes
        self.weights = weights
        self.bounds = bounds
        super().__init__()

    def integrate(self, fun, isvectorized=False, **kwargs):
        """
        Numerically integrate the function ``fun``.

        Parameters
        ----------
        fun : callable
            Function to be integrated. Signature is ``fun(x, **kwargs)``
            where ``x`` is either a float or an ``ndarray`` with
            shape (d,). If ``fun`` can be evaluated vectorized, the
            implementation expects signature ``fun(X, **kwargs)`` where
            ``X`` is is an ``ndarray`` of shape (n, d). Making use of
            vectorization is recommended wherever possible for improved
            speed of computation.
        isvectorized : bool
            Whether integrand allows vectorised evaluation (i.e. evaluation of all nodes
            at once).
        kwargs
            Key-word arguments being passed down to ``fun`` at each
            evaluation. For example (hyper)parameters.
        """
        if isvectorized is False:
            output = 0.0
            for (node, weight) in zip(self.nodes, self.weights):
                output = output + weight * fun(node, **kwargs)
        else:
            output = self.weights @ fun(self.nodes, **kwargs)
        return output



import numpy as np


class _RandomVariableList(list):
    """
    List of RandomVariables with convenient access to means, covariances, etc.

    Parameters
    ----------
    rv_list : :obj:`list` of :obj:`RandomVariable`
    """

    def __init__(self, rv_list):
        if not isinstance(rv_list, list):
            raise TypeError("RandomVariableList expects a list.")
        super().__init__(rv_list)

    def mean(self):
        return np.stack([rv.mean for rv in self])

    def cov(self):
        return np.stack([rv.cov for rv in self])

    def var(self):
        return np.stack([rv.var for rv in self])

    def std(self):
        return np.stack([rv.std for rv in self])

    def __getitem__(self, idx):
        """Make sure to wrap the result into a _RandomVariableList if necessary"""
        result = super().__getitem__(idx)
        if isinstance(result, list):
            result = _RandomVariableList(result)
        return result



"""
Abstract ODESolver class. Interface for Runge-Kutta, ODEFilter.
"""

from abc import ABC, abstractmethod
import warnings

from probnum.diffeq.odesolution import ODESolution


class ODESolver(ABC):
    """
    Interface for ODESolver.
    """

    def __init__(self, ivp):
        self.ivp = ivp
        self.num_steps = 0

    def solve(self, firststep, steprule, **kwargs):
        """
        Solve an IVP.

        Parameters
        ----------
        firststep : float
            First step for adaptive step-size rule.
        steprule : :class:`StepRule`
            Step-size selection rule, e.g. constant steps or adaptive steps.
        """
        t, current_rv = self.initialise()
        times, rvs = [t], [current_rv]
        stepsize = firststep

        while t < self.ivp.tmax:

            t_new = t + stepsize
            proposed_rv, errorest = self.step(t, t_new, current_rv, **kwargs)

            if steprule.is_accepted(stepsize, errorest):
                self.num_steps += 1
                t = t_new
                current_rv = proposed_rv
                times.append(t)
                rvs.append(current_rv)
                self.method_callback(
                    time=t_new, current_guess=proposed_rv, current_error=errorest
                )

            suggested_stepsize = self._suggest_step(stepsize, errorest, steprule)
            stepsize = min(suggested_stepsize, self.ivp.tmax - t)

        odesol = self.postprocess(times=times, rvs=rvs)
        return odesol

    def _suggest_step(self, step, errorest, steprule):
        """
        Suggests step according to steprule and warns if step is extremely small.

        Raises
        ------
        RuntimeWarning
            If suggested step is smaller than :math:`10^{-15}`.
        """
        step = steprule.suggest(step, errorest)
        if step < 1e-15:
            warnmsg = "Stepsize is num. zero (%.1e)" % step
            warnings.warn(message=warnmsg, category=RuntimeWarning)
        return step

    @abstractmethod
    def initialise(self):
        """Returns t0 and y0 (for the solver, which might be different to ivp.y0)"""
        raise NotImplementedError

    @abstractmethod
    def step(self, start, stop, current, **kwargs):
        """Every ODE solver needs a step() method that returns a new random variable and an error estimate"""
        raise NotImplementedError

    def postprocess(self, times, rvs):
        """
        Turn list of random variables into an ODE solution object and potentially do more.
        Overwrite for instance via

        >>> def postprocess(self, times, rvs):
        >>> # do something with times and rvs
        >>> odesol = super().postprocess(times, rvs)
        >>> # do something with odesol
        >>> return odesol
        """
        return ODESolution(times, rvs, self)

    def method_callback(self, time, current_guess, current_error):
        """
        Optional callback. Can be overwritten.
        Do this as soon as it is clear that the current guess is accepted, but before storing it.
        No return. For example: tune hyperparameters (sigma).
        """
        pass



"""
Interfaces for Bayesian filtering and smoothing.
"""

from abc import ABC, abstractmethod


class BayesFiltSmooth(ABC):
    """
    Bayesian filtering and smoothing.
    """

    def __init__(self, dynamod, measmod, initrv):
        self.dynamod = dynamod
        self.measmod = measmod
        self.initrv = initrv

    @abstractmethod
    def filter_step(self, start, stop, randvar, data, **kwargs):
        """
        Filter step.

        For e.g. Gaussian filters, this means a prediction step followed
        by an update step.
        """
        errormsg = (
            "filter_step(...) is not implemented for "
            + "the Bayesian filter {}.".format(type(self).__name__)
        )
        raise NotImplementedError(errormsg)

    def smoother_step(self, **kwargs):
        """Smoother step."""
        errormsg = (
            "smoother_step(...) is not implemented for "
            + "the Bayesian smoother {}.".format(type(self).__name__)
        )
        raise NotImplementedError(errormsg)

    def predict(self, start, stop, randvar, **kwargs):
        """
        Prediction step of the Bayesian filter.

        Not required for all filters, e.g. the Particle Filter only
        has an `update()` method.
        """
        classname = type(self).__name__
        errormsg = (
            "predict(...) is not implemented for "
            + "the Bayesian filter {}.".format(classname)
        )
        raise NotImplementedError(errormsg)

    def update(self, time, randvar, data, **kwargs):
        """
        Update step of the Bayesian filter.

        Must be implemented by subclasses.
        """
        raise NotImplementedError

    @property
    def dynamicmodel(self):
        """
        Convenience function for accessing ``self.dynamod``.
        """
        return self.dynamod

    @property
    def measurementmodel(self):
        """
        Convenience function for accessing ``self.measmod``.
        """
        return self.measmod

    @property
    def initialrandomvariable(self):
        """
        Convenience function for accessing ``self.initrv``.
        """
        return self.initrv

    @property
    def initialdistribution(self):
        """
        Convenience function for accessing ``self.initdist``.
        """
        return self.initrv



"""
This package implements random variables. Random variables are the primary in- and
outputs of probabilistic numerical methods. A generic signature of such methods looks
like this:

.. highlight:: python
.. code-block:: python

    randvar_out, info = probnum_method(problem, randvar_in, **kwargs)

"""

from ._random_variable import (
    RandomVariable,
    DiscreteRandomVariable,
    ContinuousRandomVariable,
)

from ._dirac import Dirac
from ._normal import Normal

from ._scipy_stats import (
    WrappedSciPyRandomVariable,
    WrappedSciPyDiscreteRandomVariable,
    WrappedSciPyContinuousRandomVariable,
)

from ._utils import asrandvar

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "asrandvar",
    "RandomVariable",
    "DiscreteRandomVariable",
    "ContinuousRandomVariable",
    "Dirac",
    "Normal",
    "WrappedSciPyRandomVariable",
    "WrappedSciPyDiscreteRandomVariable",
    "WrappedSciPyContinuousRandomVariable",
]

# Set correct module paths. Corrects links and module paths in documentation.
RandomVariable.__module__ = "probnum.random_variables"
DiscreteRandomVariable.__module__ = "probnum.random_variables"
ContinuousRandomVariable.__module__ = "probnum.random_variables"

WrappedSciPyRandomVariable.__module__ = "probnum.random_variables"
WrappedSciPyDiscreteRandomVariable.__module__ = "probnum.random_variables"
WrappedSciPyContinuousRandomVariable.__module__ = "probnum.random_variables"

Dirac.__module__ = "probnum.random_variables"
Normal.__module__ = "probnum.random_variables"

asrandvar.__module__ = "probnum.random_variables"



"""
Import convenience functions in optim.py to create an
intuitive, numpy-like interface.

Note
----
Local import, because with a global import this does not seem
to work.
"""

from .ivp2filter import *
from .ivpfiltsmooth import *
from .odefiltsmooth import *
from .prior import *



"""
Clenshaw-Curtis quadrature rule.

This module implements the Clenshaw-Curtis quadrature rule and associated functions.

Formula for nodes and weights:
    [1] Sparse Grid Quadrature in High Dimensions with Applications in Finance and
        Insurance Holtz, M., Springer, 2010(, Chapter 3, p. 42ff)
URL:
    https://books.google.de/books?id=XOfMm-4ZM9AC&pg=PA42&lpg=PA42&dq=filippi+formu
    la+clenshaw+curtis&source=bl&ots=gkhNu9F1fp&sig=ACfU3U3zdH-OHx0PqqB_KAXb1mM5iXI
    ojw&hl=de&sa=X&ved=2ahUKEwijovCC1O7kAhUKr6QKHaVWCwQQ6AEwD3oECAgQAQ#v=onepage&q=
    filippi%20formula%20clenshaw%20curtis&f=false

Note
----
This is a lot of old code. It is well tested, but the _compute_*(...)
functions are really hard to read (they are efficient, though).
"""

import numpy as np

from probnum.quad.polynomial.polynomialquadrature import PolynomialQuadrature
from probnum import utils


class ClenshawCurtis(PolynomialQuadrature):
    """
    Clenshaw-Curtis quadrature rule.

    Method of numerical integration based on an expansion of the
    integrand in terms of a discrete cosine transform.


    The nodes of the Clenshaw-Curtis rule are the roots of the Chebyshev
    polynomials. The :math:`i^\\text{th}` root is

    .. math:: x_i = \\frac{1}{2} \\left(1 - \\cos\\left( \\frac{i \\pi}{n+1} \\right) \\right)

    for :math:`i=1, ..., n`. The :math:`i^\\text{th}` weight is given by

    .. math:: w_i = \\frac{2}{n+1} \\sin\\left(\\frac{i \\pi}{n+1}\\right)\\sum_{j=1}^{(n+1)/2} \\frac{1}{2j-1}\\sin\\left(\\frac{(2j-1)i \\pi}{n+1}\\right).

    These formulas can be found in [1]_. For an :math:`r`-times
    differentiable integrand, the Clenshaw-Curtis approximation error is
    proportional to :math:`\\mathcal{O}(n^{-r})`. It integrates
    polynomials of degree :math:`\\leq n+1` exactly.

    Parameters
    ----------
    npts_per_dim : int
        Number of evaluation points per dimension. The resulting mesh
        will have `npts_per_dim**ndim` elements.
    ndim : int
        Number of dimensions.
    bounds : ndarray, shape=(n, 2)
        Integration bounds.

    See Also
    --------
    PolynomialQuadrature : Quadrature rule based on polynomial functions.

    References
    ----------
    .. [1] Holtz, M., Sparse Grid Quadrature in High Dimensions with Applications in
       Finance and Insurance, Springer, 2010


    Examples
    --------
    >>> cc = ClenshawCurtis(npts_per_dim=3, ndim=2, bounds=np.array([[0, 1], [0, 0.1]]))
    >>> print(cc.nodes)
    [[0.14644661 0.01464466]
     [0.14644661 0.05      ]
     [0.14644661 0.08535534]
     [0.5        0.01464466]
     [0.5        0.05      ]
     [0.5        0.08535534]
     [0.85355339 0.01464466]
     [0.85355339 0.05      ]
     [0.85355339 0.08535534]]
    >>> print(cc.weights)
    [0.01111111 0.01111111 0.01111111 0.01111111 0.01111111 0.01111111
     0.01111111 0.01111111 0.01111111]
    >>> print(cc.integrate(lambda x: x[0] + x[1]))
    0.05500000000000001

    >>> cc = ClenshawCurtis(npts_per_dim=7, ndim=1, bounds=np.array([[0, 1]]))
    >>> print(cc.weights)
    [0.08898234 0.12380952 0.19673195 0.18095238 0.19673195 0.12380952
     0.08898234]
    >>> print(cc.nodes)
    [[0.03806023]
     [0.14644661]
     [0.30865828]
     [0.5       ]
     [0.69134172]
     [0.85355339]
     [0.96193977]]
    >>> print(cc.integrate(lambda x: np.sin(x)))
    [0.45969769]

    """  # pylint: disable=line-too-long

    def __init__(self, npts_per_dim, ndim, bounds):
        utils.assert_is_2d_ndarray(bounds)
        weights = _compute_weights(npts_per_dim, ndim, bounds)
        nodes = _compute_nodes(npts_per_dim, ndim, bounds)
        PolynomialQuadrature.__init__(self, nodes, weights, bounds)


def _compute_weights(npts, ndim, ilbds):
    """
    Computes 1D Clenshaw-Curtis weights and aligns them in
    correspondence to the computed nodes. Since the resulting mesh is of
    size (n**d, d), the weight array is of size (n**d,).
    """
    if npts ** ndim * ndim >= 1e9:
        raise MemoryError("Weights for tensor-mesh too large for memory.")
    # num_tiles = np.arange(ndim)
    num_reps = ndim - np.arange(ndim) - 1
    weights = _compute_weights_1d(npts, ndim, ilbds[0])
    prodweights = np.repeat(weights, npts ** (num_reps[0]))
    for i in range(1, ndim):
        weights = _compute_weights_1d(npts, ndim, ilbds[i])
        column = np.repeat(
            np.tile(weights, int(npts ** i)), int(npts ** (ndim - 1 - i))
        )
        prodweights *= column
    return prodweights


def _compute_weights_1d(npts, ndim, ilbds1d):
    """
    Computes weights of Clenshaw-Curtis formula.

    The :math:`i^\textrm{th}` weight is given by

    .. math:: w_i = \\frac{2}{n+1} \\sin\\left(\\frac{i \\pi}{n+1}\\right)\\sum_{j=1}^{(n+1)/2} \\frac{1}{2j-1}\\sin\\left(\\frac{(2j-1)i \\pi}{n+1}\\right).
    """  # pylint: disable=line-too-long
    if npts % 2 == 0:
        raise ValueError("Please enter odd npts")
    nhalfpts = int((npts + 1.0) / 2.0)
    ind_j = 2.0 * np.arange(1, nhalfpts + 1) - 1.0
    ind_i = np.arange(1, npts + 1)
    arr1 = 2.0 / (npts + 1.0) * np.sin(ind_i * np.pi / (npts + 1.0))
    arr2 = 1.0 / ind_j
    arr3 = np.sin(np.outer(ind_j, ind_i) * np.pi / (npts + 1.0))
    weights = arr1 * (arr2 @ arr3)
    return (ilbds1d[1] - ilbds1d[0]) * weights


def _compute_nodes(npts, ndim, ilbds):
    """
    Computes 1D Clenshaw-Curtis nodes and aligns them in order to create
    a tensor mesh: each point is aligned with each point to create a
    mesh of size (n^d, d).
    """
    if npts ** ndim * ndim >= 1e9:
        raise ValueError("Tensor-mesh too large for memory.")
    nodes = _compute_nodes_1d(npts, ilbds[0])
    productmesh = np.repeat(nodes, npts ** (ndim - 1))
    for i in range(1, ndim):
        nodes = _compute_nodes_1d(npts, ilbds[i])
        column = np.repeat(np.tile(nodes, int(npts ** i)), int(npts ** (ndim - 1 - i)))
        productmesh = np.vstack((productmesh.T, column)).T
    if ndim == 1:
        return productmesh.reshape((npts, 1))
    else:
        return productmesh


def _compute_nodes_1d(npts, ilbds1d):
    """
    Computes Clenshaw-Curtis nodes in 1d.

    The :math:`i^\\text{th}` root is

    .. math:: x_i = \\frac{1}{2} \\left(1 - \\cos\\left( \\frac{i \\pi}{n+1} \\right) \\right)

    Parameters
    ----------
    npts : int
        number of 1d quadrature points
    ilbds1d : np.ndarray of shape (2,)
        integration bounds of the form [lower, upper]

    Raises
    ------
    ValueError
        if  'npts' is even (CC needs odd number of points for reasons)

    Returns
    -------
    np.ndarray, shape (npts,)
        1d CC nodes in ilbds1d
    """  # pylint: disable=line-too-long
    if npts % 2 == 0:
        raise ValueError("Please enter odd npts")
    ind = np.arange(1, npts + 1)
    nodes = 0.5 * (1 - np.cos(np.pi * ind / (npts + 1)))
    return nodes * (ilbds1d[1] - ilbds1d[0]) + ilbds1d[0]



"""
Utility functions for functions, methods and the like.
"""

import numpy as np


__all__ = ["assert_evaluates_to_scalar"]


def assert_evaluates_to_scalar(fct, valid_input):
    """
    Checks whether the output of a function is a scalar.
    """
    if not np.isscalar(fct(valid_input)):
        raise ValueError("Function does not evaluate to scalar.")



"""
ProbNum
=====
Probabilistic Numerics in Python.
  1. Linear Algebra
  2. Quadrature
  3. Differential Equations
  4. Bayesian filtering and smoothing
and more.


Available subpackages
---------------------
prob
    Random variables, distributions and sampling.
linalg
    Probabilistic numerical linear algebra.
quad
    Bayesian quadrature / numerical integration.
filtsmooth
    Bayesian filtering and smoothing.
diffeq
    Probabilistic solvers for ordinary differential equations.
"""
# pylint: disable=wrong-import-order

from . import diffeq
from . import filtsmooth
from . import linalg
from . import quad
from . import random_variables
from . import utils

from .random_variables import asrandvar, RandomVariable

# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound



"""
Operators of Kronecker-type or related.

This module implements operators of Kronecker-type or linked to Kronecker-type products.
"""
import numpy as np

from probnum.linalg.linops.linearoperators import LinearOperator, aslinop


class Symmetrize(LinearOperator):
    """
    Symmetrizes a vector in its matrix representation.

    Given a vector x=vec(X) representing a square matrix X, this linear operator
    computes y=vec(1/2(X + X^T)).

    Parameters
    ----------
    dim : int
        Dimension of matrix X.
    """

    def __init__(self, dim):
        self._dim = dim
        super().__init__(dtype=float, shape=(dim * dim, dim * dim))

    def _matvec(self, x):
        """Assumes x=vec(X)."""
        X = np.reshape(x.copy(), (self._dim, self._dim))
        Y = 0.5 * (X + X.T)
        return Y.reshape(-1, 1)


class Vec(LinearOperator):
    """
    Vectorization operator.

    The column- or row-wise vectorization operator stacking the columns or rows of a
    matrix representation of a linear operator into a vector.

    Parameters
    ----------
    order : str
        Stacking order to apply. One of ``row`` or ``col``. Defaults to column-wise
        stacking.
    dim : int
        Either number of rows or columns, depending on the vectorization ``order``.
    """

    def __init__(self, order="col", dim=None):
        if order not in ["col", "row"]:
            raise ValueError(
                "Not a valid stacking order. Choose one of 'col' or 'row'."
            )
        self.mode = order
        super().__init__(dtype=float, shape=(dim, dim))

    def _matvec(self, x):
        """Vectorization of a vector is the identity."""
        return x

    def _matmat(self, X):
        """Stacking of matrix rows or columns."""
        if self.mode == "row":
            return X.ravel(order="C")
        else:
            return X.ravel(order="F")


class Svec(LinearOperator):
    """
    Symmetric vectorization operator.

    The column- or row-wise symmetric normalized vectorization operator
    :math:`\\operatorname{svec}` [1]_ stacking the (normalized) lower/upper triangular
    components of a symmetric matrix of a linear operator into a vector. It is defined
    by

    .. math::
        \\operatorname{svec}(S) = \\begin{bmatrix}
                                    S_{11} \\\\
                                    \\sqrt{2} S_{21} \\\\
                                    \\vdots \\\\
                                    \\sqrt{2} S_{n1} \\\\
                                    S_{22} \\\\
                                    \\sqrt{2} S_{32} \\\\
                                    \\vdots \\\\
                                    \\sqrt{2} S_{n2} \\\\
                                    \\vdots \\\\
                                    S_{nn}
                                  \\end{bmatrix}

    where :math:`S` is a symmetric linear operator defined on :math:`\\mathbb{R}^n`.

    Parameters
    ----------
    dim : int
        Dimension of the symmetric matrix to be reshaped.
    check_symmetric : bool, default=False
        Check whether the given matrix or vector corresponds to a symmetric matrix
        argument. Note, this option can slow down performance.

    Notes
    -----
    It holds that :math:`Q\\operatorname{svec}(S) = \\operatorname{vec}(S)`, where
    :math:`Q` is a unique matrix with orthonormal rows.

    References
    ----------
    .. [1] De Klerk, E., Aspects of Semidefinite Programming, *Kluwer Academic
       Publishers*, 2002
    """

    def __init__(self, dim, check_symmetric=False):
        if not isinstance(dim, int) or dim <= 0:
            raise ValueError(
                "Dimension of the input matrix S must be a positive integer."
            )
        self._dim = dim
        self.check_symmetric = check_symmetric
        super().__init__(dtype=float, shape=(dim * dim, int(0.5 * dim * (dim + 1))))

    def _matvec(self, x):
        """Assumes x = vec(X)."""
        X = np.reshape(x.copy(), (self._dim, self._dim))
        if self.check_symmetric and not (X.T == X).all():
            raise ValueError(
                "The given vector does not correspond to a symmetric matrix."
            )

        X[np.triu_indices(self._dim, k=1)] *= np.sqrt(2)
        ind = np.triu_indices(self._dim, k=0)
        return X[ind]

    def _matmat(self, X):
        """
        Vectorizes X if of dimension n^2, otherwise applies Svec to each column of X.
        """
        if np.shape(X)[0] == np.shape(X)[1] == self._dim:
            return self._matvec(X.ravel())
        elif np.shape(X)[0] == self._dim * self._dim:
            return np.hstack([self._matvec(col.reshape(-1, 1)) for col in X.T])
        else:
            raise ValueError(
                "Dimension mismatch. Argument must be either a (n x n) matrix or "
                "(n^2 x k)"
            )


class Kronecker(LinearOperator):
    """
    Kronecker product of two linear operators.

    The Kronecker product [1]_ :math:`A \\otimes B` of two linear operators :math:`A`
    and :math:`B` is given by

    .. math::
        A \\otimes B = \\begin{bmatrix}
            A_{11} B   &  \\dots   & A_{1 m_1} B  \\\\
            \\vdots     &  \\ddots  & \\vdots \\\\
            A_{n_11} B &  \\dots   & A_{n_1 m_1} B
        \\end{bmatrix}

    where :math:`A_{ij}v=A(v_j e_i)`, where :math:`e_i` is the :math:`i^{\\text{th}}`
    unit vector. The result is a new linear operator mapping from
    :math:`\\mathbb{R}^{n_1n_2}` to :math:`\\mathbb{R}^{m_1m_2}`. By recognizing that
    :math:`(A \\otimes B)\\operatorname{vec}(X) = AXB^{\\top}`, the Kronecker product
    can be understood as "translation" between matrix multiplication and (row-wise)
    vectorization.

    Parameters
    ----------
    A : np.ndarray or LinearOperator
        The first operator.
    B : np.ndarray or LinearOperator
        The second operator.
    dtype : dtype
        Data type of the operator.

    References
    ----------
    .. [1] Van Loan, C. F., The ubiquitous Kronecker product, *Journal of Computational
        and Applied Mathematics*, 2000, 123, 85-100

    See Also
    --------
    SymmetricKronecker : The symmetric Kronecker product of two linear operators.

    """

    # todo: extend this to list of operators
    def __init__(self, A, B, dtype=None):
        self.A = aslinop(A)
        self.B = aslinop(B)
        super().__init__(
            dtype=dtype,
            shape=(
                self.A.shape[0] * self.B.shape[0],
                self.A.shape[1] * self.B.shape[1],
            ),
        )

    def _matvec(self, X):
        """
        Efficient multiplication via (A (x) B)vec(X) = vec(AXB^T) where vec is the
        row-wise vectorization operator.
        """
        X = X.reshape(self.A.shape[1], self.B.shape[1])
        Y = self.B.matmat(X.T)
        return self.A.matmat(Y.T).ravel()

    def _rmatvec(self, X):
        # (A (x) B)^T = A^T (x) B^T.
        X = X.reshape(self.A.shape[0], self.B.shape[0])
        Y = self.B.H.matmat(X.T)
        return self.A.H.matmat(Y.T).ravel()

    def transpose(self):
        # (A (x) B)^T = A^T (x) B^T
        return Kronecker(A=self.A.transpose(), B=self.B.transpose(), dtype=self.dtype)

    def inv(self):
        #  (A (x) B)^-1 = A^-1 (x) B^-1
        return Kronecker(A=self.A.inv(), B=self.B.inv(), dtype=self.dtype)

    # Properties
    def rank(self):
        return self.A.rank() * self.B.rank()

    def eigvals(self):
        raise NotImplementedError

    def cond(self, p=None):
        return self.A.cond(p=p) * self.B.cond(p=p)

    def det(self):
        # If A (m x m) and B (n x n), then det(A (x) B) = det(A)^n * det(B) * m
        if self.A.shape[0] == self.A.shape[1] and self.B.shape[0] == self.B.shape[1]:
            return self.A.det() ** self.B.shape[0] * self.B.det() ** self.A.shape[0]
        else:
            raise NotImplementedError

    def logabsdet(self):
        # If A (m x m) and B (n x n), then det(A (x) B) = det(A)^n * det(B) * m
        if self.A.shape[0] == self.A.shape[1] and self.B.shape[0] == self.B.shape[1]:
            return (
                self.B.shape[0] * self.A.logabsdet()
                + self.A.shape[0] * self.B.logabsdet()
            )
        else:
            raise NotImplementedError

    def trace(self):
        if self.A.shape[0] == self.A.shape[1] and self.B.shape[0] == self.B.shape[1]:
            return self.A.trace() * self.B.trace()
        else:
            raise NotImplementedError


class SymmetricKronecker(LinearOperator):
    """
    Symmetric Kronecker product of two linear operators.

    The symmetric Kronecker product [1]_ :math:`A \\otimes_{s} B` of two square linear
    operators :math:`A` and :math:`B` maps a symmetric linear operator :math:`X` to
    :math:`\\mathbb{R}^{\\frac{1}{2}n (n+1)}`. It is given by

    .. math::
        (A \\otimes_{s} B)\\operatorname{svec}(X) = \\frac{1}{2} \\operatorname{svec}(AXB^{\\top} + BXA^{\\top})

    where :math:`\\operatorname{svec}(X) = (X_{11}, \\sqrt{2} X_{12}, \\dots, X_{1n},
    X_{22}, \\sqrt{2} X_{23}, \\dots, \\sqrt{2}X_{2n}, \\dots X_{nn})^{\\top}` is the
    (row-wise, normalized) symmetric stacking operator. The implementation is based on
    the relationship :math:`Q^\\top \\operatorname{svec}(X) = \\operatorname{vec}(X)`
    with an orthonormal matrix :math:`Q` [2]_.

    Note
    ----
    The symmetric Kronecker product has a symmetric matrix representation if both
    :math:`A` and :math:`B` are symmetric.

    References
    ----------
    .. [1] Van Loan, C. F., The ubiquitous Kronecker product, *Journal of Computational
       and Applied Mathematics*, 2000, 123, 85-100
    .. [2] De Klerk, E., Aspects of Semidefinite Programming, *Kluwer Academic
       Publishers*, 2002

    See Also
    --------
    Kronecker : The Kronecker product of two linear operators.
    """  # pylint: disable=line-too-long

    # TODO: update documentation to map from n2xn2 to matrices of rank 1/2n(n+1),
    # representation symmetric n2xn2

    def __init__(self, A, B=None, dtype=None):
        # Set parameters
        self.A = aslinop(A)
        self._ABequal = False
        if B is None:
            self.B = self.A
            self._ABequal = True
        else:
            self.B = aslinop(B)
        self._n = self.A.shape[0]
        if self.A.shape != self.B.shape or self.A.shape[1] != self._n:
            raise ValueError(
                "Linear operators A and B must be square and have the same dimensions."
            )

        # Initiator of superclass
        super().__init__(dtype=dtype, shape=(self._n ** 2, self._n ** 2))

    def _matvec(self, x):
        """
        Efficient multiplication via (A (x)_s B)vec(X) = 1/2 vec(BXA^T + AXB^T) where
        vec is the column-wise normalized symmetric stacking operator.
        """
        # vec(x)
        X = x.reshape(self._n, self._n)

        # (A (x)_s B)vec(X) = 1/2 vec(BXA^T + AXB^T)
        Y1 = (self.A @ (self.B @ X).T).T
        Y2 = (self.B @ (self.A @ X).T).T
        Y = 0.5 * (Y1 + Y2)
        return Y.ravel()

    def _rmatvec(self, x):
        """Based on (A (x)_s B)^T = A^T (x)_s B^T."""
        # vec(x)
        X = x.reshape(self._n, self._n)

        # (A^T (x)_s B^T)vec(X) = 1/2 vec(B^T XA + A^T XB)
        Y1 = (self.A.H @ (self.B.H @ X).T).T
        Y2 = (self.B.H @ (self.A.H @ X).T).T
        Y = 0.5 * (Y1 + Y2)
        return Y.ravel()

    # TODO: add efficient implementation of _matmat based on (Symmetric) Kronecker
    # properties

    def todense(self):
        """Dense representation of the symmetric Kronecker product"""
        # 1/2 (A (x) B + B (x) A)
        A_dense = self.A.todense()
        B_dense = self.B.todense()
        return 0.5 * (np.kron(A_dense, B_dense) + np.kron(B_dense, A_dense))

    def inv(self):
        # (A (x)_s A)^-1 = A^-1 (x)_s A^-1
        if self._ABequal:
            return SymmetricKronecker(A=self.A.inv(), dtype=self.dtype)
        else:
            return NotImplementedError



"""
Linear Algebra.

This package implements common operations and (probabilistic) numerical methods for
linear algebra.
"""

from probnum.linalg.linearsolvers import *

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "problinsolve",
    "bayescg",
    "ProbabilisticLinearSolver",
    "MatrixBasedSolver",
    "AsymmetricMatrixBasedSolver",
    "SymmetricMatrixBasedSolver",
    "SolutionBasedSolver",
]

# Set correct module paths. Corrects links and module paths in documentation.
ProbabilisticLinearSolver.__module__ = "probnum.linalg"
MatrixBasedSolver.__module__ = "probnum.linalg"



"""
Gaussian filtering.
"""
from abc import ABC, abstractmethod

import numpy as np

from probnum.random_variables import Normal
from probnum.filtsmooth.bayesfiltsmooth import BayesFiltSmooth
from probnum.filtsmooth.gaussfiltsmooth.kalmanposterior import KalmanPosterior


class GaussFiltSmooth(BayesFiltSmooth, ABC):
    """
    Interface for Gaussian filtering and smoothing.
    """

    def __init__(self, dynamod, measmod, initrv):
        """Check that the initial distribution is Gaussian."""
        if not issubclass(type(initrv), Normal):
            raise ValueError(
                "Gaussian filters/smoothers need initial "
                "random variables with Normal distribution."
            )
        super().__init__(dynamod, measmod, initrv)

    def filtsmooth(self, dataset, times, **kwargs):
        """
        Apply Gaussian filtering and smoothing to a data set.

        Parameters
        ----------
        dataset : array_like, shape (N, M)
            Data set that is filtered.
        times : array_like, shape (N,)
            Temporal locations of the data points.
        kwargs : ???

        Returns
        -------
        KalmanPosterior
            Posterior distribution of the smoothed output
        """
        dataset, times = np.asarray(dataset), np.asarray(times)
        filter_posterior = self.filter(dataset, times, **kwargs)
        smooth_posterior = self.smooth(filter_posterior, **kwargs)
        return smooth_posterior

    def filter(self, dataset, times, **kwargs):
        """
        Apply Gaussian filtering (no smoothing!) to a data set.

        Parameters
        ----------
        dataset : array_like, shape (N, M)
            Data set that is filtered.
        times : array_like, shape (N,)
            Temporal locations of the data points.
        kwargs : ???

        Returns
        -------
        KalmanPosterior
            Posterior distribution of the filtered output
        """
        dataset, times = np.asarray(dataset), np.asarray(times)
        filtrv = self.initialrandomvariable
        rvs = [filtrv]
        for idx in range(1, len(times)):
            filtrv = self.filter_step(
                start=times[idx - 1],
                stop=times[idx],
                randvar=filtrv,
                data=dataset[idx - 1],
                **kwargs
            )
            rvs.append(filtrv)
        return KalmanPosterior(times, rvs, self)

    def smooth(self, filter_posterior, **kwargs):
        """
        Apply Gaussian smoothing to a set of filtered means and covariances.

        Parameters
        ----------
        filter_posterior : KalmanPosterior
            Posterior distribution obtained after filtering
        kwargs : ???

        Returns
        -------
        KalmanPosterior
            Posterior distribution of the smoothed output
        """
        smoothed_rv = filter_posterior[-1]
        locations = filter_posterior.locations
        out_rvs = [smoothed_rv]
        for idx in reversed(range(1, len(locations))):
            unsmoothed_rv = filter_posterior[idx - 1]
            pred_rv, ccov = self.predict(
                start=locations[idx - 1],
                stop=locations[idx],
                randvar=unsmoothed_rv,
                **kwargs
            )
            smoothed_rv = self.smooth_step(unsmoothed_rv, pred_rv, smoothed_rv, ccov)
            out_rvs.append(smoothed_rv)
        out_rvs.reverse()
        return KalmanPosterior(locations, out_rvs, self)

    def filter_step(self, start, stop, randvar, data, **kwargs):
        """
        A single filter step.

        Consists of a prediction step (t -> t+1) and an update step (at t+1).

        Parameters
        ----------
        start : float
            Predict FROM this time point.
        stop : float
            Predict TO this time point.
        randvar : RandomVariable
            Predict based on this random variable. For instance, this can be the result
            of a previous call to filter_step.
        data : array_like
            Compute the update based on this data.

        Returns
        -------
        RandomVariable
            Resulting filter estimate after the single step.
        """
        data = np.asarray(data)
        predrv, _ = self.predict(start, stop, randvar, **kwargs)
        filtrv, _, _, _ = self.update(stop, predrv, data)
        return filtrv

    def smooth_step(self, unsmoothed_rv, pred_rv, smoothed_rv, crosscov):
        """
        A single smoother step.

        Consists of predicting from the filtering distribution at time t
        to time t+1 and then updating based on the discrepancy to the
        smoothing solution at time t+1.

        Parameters
        ----------
        unsmoothed_rv : RandomVariable
            Filtering distribution at time t.
        pred_rv : RandomVariable
            Prediction at time t+1 of the filtering distribution at time t.
        smoothed_rv : RandomVariable
            Smoothing distribution at time t+1.
        crosscov : array_like
            Cross-covariance between unsmoothed_rv and pred_rv as
            returned by predict().
        """
        crosscov = np.asarray(crosscov)
        initmean, initcov = unsmoothed_rv.mean, unsmoothed_rv.cov
        predmean, predcov = pred_rv.mean, pred_rv.cov
        currmean, currcov = smoothed_rv.mean, smoothed_rv.cov
        if np.isscalar(predmean) and np.isscalar(predcov):
            predmean = predmean * np.ones(1)
            predcov = predcov * np.eye(1)
        newmean = initmean + crosscov @ np.linalg.solve(predcov, currmean - predmean)
        firstsolve = crosscov @ np.linalg.solve(predcov, currcov - predcov)
        secondsolve = crosscov @ np.linalg.solve(predcov, firstsolve.T)
        newcov = initcov + secondsolve.T
        return Normal(newmean, newcov)

    @abstractmethod
    def predict(self, start, stop, randvar, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def update(self, time, randvar, data, **kwargs):
        raise NotImplementedError


def linear_discrete_update(meanest, cpred, data, meascov, measmat, mpred):
    """Kalman update, potentially after linearization."""
    covest = measmat @ cpred @ measmat.T + meascov
    ccest = cpred @ measmat.T
    mean = mpred + ccest @ np.linalg.solve(covest, data - meanest)
    cov = cpred - ccest @ np.linalg.solve(covest.T, ccest.T)
    return Normal(mean, cov), covest, ccest, meanest



import numpy as np
import scipy.sparse

from . import _random_variable
from . import _dirac
from . import _scipy_stats


def asrandvar(obj) -> _random_variable.RandomVariable:
    """
    Return ``obj`` as a :class:`RandomVariable`.

    Converts scalars, (sparse) arrays or distribution classes to a
    :class:`RandomVariable`.

    Parameters
    ----------
    obj : object
        Argument to be represented as a :class:`RandomVariable`.

    Returns
    -------
    rv : RandomVariable
        The object `obj` as a :class:`RandomVariable`.

    See Also
    --------
    RandomVariable : Class representing random variables.

    Examples
    --------
    >>> from scipy.stats import bernoulli
    >>> from probnum import asrandvar
    >>> bern = bernoulli(p=0.5)
    >>> bern.random_state = 42  # Seed for reproducibility
    >>> b = asrandvar(bern)
    >>> b.sample(size=5)
    array([1, 1, 1, 0, 0])
    """

    # RandomVariable
    if isinstance(obj, _random_variable.RandomVariable):
        return obj
    # Scalar
    elif np.isscalar(obj):
        return _dirac.Dirac(support=obj)
    # Numpy array, sparse array or Linear Operator
    elif isinstance(
        obj, (np.ndarray, scipy.sparse.spmatrix, scipy.sparse.linalg.LinearOperator)
    ):
        return _dirac.Dirac(support=obj)
    # Scipy random variable
    elif isinstance(
        obj,
        (
            scipy.stats._distn_infrastructure.rv_frozen,
            scipy.stats._multivariate.multi_rv_frozen,
        ),
    ):
        return _scipy_stats.wrap_scipy_rv(obj)
    else:
        raise ValueError(
            f"Argument of type {type(obj)} cannot be converted to a random variable."
        )



"""
Matrix-based probabilistic linear solvers.

Implementations of matrix-based linear solvers which perform inference on the matrix or
its inverse given linear observations.
"""
import warnings
import abc

import numpy as np
import GPy

import probnum
from probnum import random_variables as rvs
from probnum.linalg import linops


class ProbabilisticLinearSolver(abc.ABC):
    """
    An abstract base class for probabilistic linear solvers.

    This class is designed to be subclassed with new (probabilistic) linear solvers,
    which implement a ``.solve()`` method. Objects of this type are instantiated in
    wrapper functions such as :meth:``problinsolve``.

    Parameters
    ----------
    A : array-like or LinearOperator or RandomVariable, shape=(n,n)
        A square matrix or linear operator. A prior distribution can be provided as a
        :class:`~probnum.RandomVariable`. If an array or linear operator is given,
        a prior distribution is chosen automatically.
    b : RandomVariable, shape=(n,) or (n, nrhs)
        Right-hand side vector, matrix or RandomVariable of :math:`A x = b`.
    """

    def __init__(self, A, b):
        self.A = A
        self.b = b
        self.n = A.shape[1]

    def has_converged(self, iter, maxiter, **kwargs):
        """
        Check convergence of a linear solver.

        Evaluates a set of convergence criteria based on its input arguments to decide
        whether the iteration has converged.

        Parameters
        ----------
        iter : int
            Current iteration of solver.
        maxiter : int
            Maximum number of iterations

        Returns
        -------
        has_converged : bool
            True if the method has converged.
        convergence_criterion : str
            Convergence criterion which caused termination.
        """
        # maximum iterations
        if iter >= maxiter:
            warnings.warn(
                "Iteration terminated. Solver reached the maximum number of iterations."
            )
            return True, "maxiter"
        else:
            return False, ""

    def solve(self, callback=None, **kwargs):
        """
        Solve the linear system :math:`Ax=b`.

        Parameters
        ----------
        callback : function, optional
            User-supplied function called after each iteration of the linear solver. It
            is called as ``callback(xk, Ak, Ainvk, sk, yk, alphak, resid, **kwargs)``
            and can be used to return quantities from the iteration. Note that depending
            on the function supplied, this can slow down the solver.
        kwargs
            Key-word arguments adjusting the behaviour of the ``solve`` iteration. These
            are usually convergence criteria.

        Returns
        -------
        x : RandomVariable, shape=(n,) or (n, nrhs)
            Approximate solution :math:`x` to the linear system. Shape of the return
            matches the shape of ``b``.
        A : RandomVariable, shape=(n,n)
            Posterior belief over the linear operator.
        Ainv : RandomVariable, shape=(n,n)
            Posterior belief over the linear operator inverse :math:`H=A^{-1}`.
        info : dict
            Information on convergence of the solver.

        """
        raise NotImplementedError


class MatrixBasedSolver(ProbabilisticLinearSolver, abc.ABC):
    """
    Abstract class for matrix-based probabilistic linear solvers.

    Parameters
    ----------
    A : array-like or LinearOperator or RandomVariable, shape=(n,n)
        A square matrix or linear operator. A prior distribution can be provided as a
        :class:`~probnum.RandomVariable`. If an array or linear operator is given,
        a prior distribution is
        chosen automatically.
    b : RandomVariable, shape=(n,) or (n, nrhs)
        Right-hand side vector, matrix or RandomVariable of :math:`A x = b`.
    x0 : array-like, shape=(n,) or (n, nrhs)
        Optional. Guess for the solution of the linear system.
    """

    def __init__(self, A, b, x0=None):
        self.x0 = x0
        super().__init__(A=A, b=b)

    def _get_prior_params(self, A0, Ainv0, x0, b):
        """
        Parameters
        ----------
        A0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
            A square matrix, linear operator or random variable representing the prior
            belief over the linear operator :math:`A`. If an array or linear operator is
            given, a prior distribution is chosen automatically.
        Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
            A square matrix, linear operator or random variable representing the prior
            belief over the inverse :math:`H=A^{-1}`. This can be viewed as taking the
            form of a pre-conditioner. If an array or linear operator is given, a prior
            distribution is chosen automatically.
        x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
            Optional. Prior belief for the solution of the linear system. Will be
            ignored if ``A0`` or ``Ainv0`` is given.
        b : array_like, shape=(n,) or (n, nrhs)
            Right-hand side vector or matrix in :math:`A x = b`.
        """
        raise NotImplementedError

    def _construct_symmetric_matrix_prior_means(self, A, x0, b):
        """
        Create matrix prior means from an initial guess for the solution of the linear
        system.

        Constructs a matrix-variate prior mean for H from ``x0`` and ``b`` such that
        :math:`H_0b = x_0`, :math:`H_0` symmetric positive definite and
        :math:`A_0 = H_0^{-1}`.

        Parameters
        ----------
        A : array-like or LinearOperator, shape=(n,n)
            System matrix assumed to be square.
        x0 : array-like, shape=(n,) or (n, nrhs)
            Optional. Guess for the solution of the linear system.
        b : array_like, shape=(n,) or (n, nrhs)
            Right-hand side vector or matrix in :math:`A x = b`.

        Returns
        -------
        A0_mean : linops.LinearOperator
            Mean of the matrix-variate prior distribution on the system matrix
            :math:`A`.
        Ainv0_mean : linops.LinearOperator
            Mean of the matrix-variate prior distribution on the inverse of the system
            matrix :math:`H = A^{-1}`.
        """
        # Check inner product between x0 and b; if negative or zero, choose better
        # initialization
        bx0 = np.squeeze(b.T @ x0)
        bb = np.linalg.norm(b) ** 2
        if bx0 < 0:
            x0 = -x0
            bx0 = -bx0
            print("Better initialization found, setting x0 = - x0.")
        elif bx0 == 0:
            if np.all(b == np.zeros_like(b)):
                print("Right-hand-side is zero. Initializing with solution x0 = 0.")
                x0 = b
            else:
                print("Better initialization found, setting x0 = (b'b/b'Ab) * b.")
                bAb = np.squeeze(b.T @ (A @ b))
                x0 = bb / bAb * b
                bx0 = bb ** 2 / bAb

        # Construct prior mean of A and H
        alpha = 0.5 * bx0 / bb

        def _mv(v):
            return (x0 - alpha * b) * (x0 - alpha * b).T @ v

        def _mm(M):
            return (x0 - alpha * b) @ (x0 - alpha * b).T @ M

        Ainv0_mean = linops.ScalarMult(
            scalar=alpha, shape=(self.n, self.n)
        ) + 2 / bx0 * linops.LinearOperator(
            matvec=_mv, matmat=_mm, shape=(self.n, self.n)
        )
        A0_mean = linops.ScalarMult(scalar=1 / alpha, shape=(self.n, self.n)) - 1 / (
            alpha * np.squeeze((x0 - alpha * b).T @ x0)
        ) * linops.LinearOperator(matvec=_mv, matmat=_mm, shape=(self.n, self.n))
        return A0_mean, Ainv0_mean

    def has_converged(self, iter, maxiter, **kwargs):
        raise NotImplementedError

    def solve(self, callback=None, maxiter=None, atol=None):
        raise NotImplementedError


class AsymmetricMatrixBasedSolver(ProbabilisticLinearSolver):
    """
    Asymmetric matrix-based probabilistic linear solver.

    Parameters
    ----------
    A : array-like or LinearOperator or RandomVariable, shape=(n,n)
        The square matrix or linear operator of the linear system.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    """

    def __init__(self, A, b, x0):
        self.x0 = x0
        super().__init__(A=A, b=b)

    def has_converged(self, iter, maxiter, **kwargs):
        raise NotImplementedError

    def solve(self, callback=None, maxiter=None, atol=None):
        raise NotImplementedError


class SymmetricMatrixBasedSolver(MatrixBasedSolver):
    """
    Symmetric matrix-based probabilistic linear solver.

    Implements the solve iteration of the symmetric matrix-based probabilistic linear
    solver described in [1]_ and [2]_.

    Parameters
    ----------
    A : array-like or LinearOperator or RandomVariable, shape=(n,n)
        The square matrix or linear operator of the linear system.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    A0 : array-like or LinearOperator or RandomVariable, shape=(n, n), optional
        A square matrix, linear operator or random variable representing the prior
        belief over the linear operator :math:`A`. If an array or linear operator is
        given, a prior distribution is chosen automatically.
    Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior
        belief over the inverse :math:`H=A^{-1}`. This can be viewed as taking the form
        of a pre-conditioner. If an array or linear operator is given, a prior
        distribution is chosen automatically.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if
        ``Ainv0`` is given.

    Returns
    -------
    A : RandomVariable
        Posterior belief over the linear operator.
    Ainv : RandomVariable
        Posterior belief over the inverse linear operator.
    x : RandomVariable
        Posterior belief over the solution of the linear system.
    info : dict
        Information about convergence and the solution.

    References
    ----------
    .. [1] Wenger, J. and Hennig, P., Probabilistic Linear Solvers for Machine Learning,
       2020
    .. [2] Hennig, P., Probabilistic Interpretation of Linear Solvers, *SIAM Journal on
       Optimization*, 2015, 25, 234-260

    See Also
    --------
    NoisySymmetricMatrixBasedSolver :
        Class implementing the noisy symmetric probabilistic linear solver.
    """

    def __init__(self, A, b, A0=None, Ainv0=None, x0=None):

        # Assume constant right hand side
        if isinstance(b, probnum.RandomVariable):
            _b = b.sample(size=1)
        else:
            _b = b

        super().__init__(A=A, b=_b, x0=x0)

        # Get or construct prior parameters
        A_mean0, A_covfactor0, Ainv_mean0, Ainv_covfactor0 = self._get_prior_params(
            A0=A0, Ainv0=Ainv0, x0=self.x0, b=self.b
        )

        # Initialize prior parameters
        self.A_mean = A_mean0
        self.A_mean0 = A_mean0
        self.A_covfactor = A_covfactor0
        self.A_covfactor0 = A_covfactor0
        self.Ainv_mean = Ainv_mean0
        self.Ainv_mean0 = Ainv_mean0
        self.Ainv_covfactor = Ainv_covfactor0
        self.Ainv_covfactor0 = Ainv_covfactor0
        if isinstance(x0, np.ndarray):
            self.x_mean = x0
        elif x0 is None:
            self.x_mean = Ainv_mean0 @ self.b
        else:
            raise NotImplementedError
        self.x0 = self.x_mean

        # Computed search directions and observations
        self.search_dir_list = []
        self.obs_list = []
        self.sy = []

    def _get_prior_params(self, A0, Ainv0, x0, b):
        """
        Get the parameters of the matrix priors on A and H.

        Retrieves and / or initializes prior parameters of ``A0`` and ``Ainv0``.

        Parameters
        ----------
        A0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
            A square matrix, linear operator or random variable representing the prior
            belief over the linear operator :math:`A`. If an array or linear operator is
            given, a prior distribution is chosen automatically. Ainv0 : array-like or
            LinearOperator or RandomVariable, shape=(n,n), optional A square matrix,
            linear operator or random variable representing the prior belief over the
            inverse :math:`H=A^{-1}`. This can be viewed as taking the form of a
            pre-conditioner. If an array or linear operator is given, a prior
            distribution is chosen automatically.
        x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
            Optional. Prior belief for the solution of the linear system. Will be
            ignored if ``A0`` or ``Ainv0`` is given.
        b : array_like, shape=(n,) or (n, nrhs)
            Right-hand side vector or matrix in :math:`A x = b`.

        Returns
        -------
        A0_mean : array-like or LinearOperator, shape=(n,n)
            Prior mean of the linear operator :math:`A`.
        A0_covfactor : array-like or LinearOperator, shape=(n,n)
            Factor :math:`W^A` of the symmetric Kronecker product prior covariance
            :math:`W^A \\otimes_s W^A` of :math:`A`.
        Ainv0_mean : array-like or LinearOperator, shape=(n,n)
            Prior mean of the linear operator :math:`H`.
        Ainv0_covfactor : array-like or LinearOperator, shape=(n,n)
            Factor :math:`W^H` of the symmetric Kronecker product prior covariance
            :math:`W^H \\otimes_s W^H` of :math:`H`.
        """
        self.is_calib_covclass = False
        # No matrix priors specified
        if A0 is None and Ainv0 is None:
            self.is_calib_covclass = True
            # No prior information given
            if x0 is None:
                Ainv0_mean = linops.Identity(shape=self.n)
                Ainv0_covfactor = linops.Identity(shape=self.n)
                # Symmetric posterior correspondence
                A0_mean = linops.Identity(shape=self.n)
                A0_covfactor = self.A
                return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor
            # Construct matrix priors from initial guess x0
            elif isinstance(x0, np.ndarray):
                A0_mean, Ainv0_mean = self._construct_symmetric_matrix_prior_means(
                    A=self.A, x0=x0, b=b
                )
                Ainv0_covfactor = Ainv0_mean
                # Symmetric posterior correspondence
                A0_covfactor = self.A
                return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor
            elif isinstance(x0, probnum.RandomVariable):
                raise NotImplementedError

        # Prior on Ainv specified
        if not isinstance(A0, probnum.RandomVariable) and Ainv0 is not None:
            if isinstance(Ainv0, probnum.RandomVariable):
                Ainv0_mean = Ainv0.mean
                Ainv0_covfactor = Ainv0.cov.A
            else:
                self.is_calib_covclass = True
                Ainv0_mean = Ainv0
                Ainv0_covfactor = Ainv0  # Symmetric posterior correspondence
            try:
                if A0 is not None:
                    A0_mean = A0
                elif isinstance(Ainv0, probnum.RandomVariable):
                    A0_mean = Ainv0.mean.inv()
                else:
                    A0_mean = Ainv0.inv()
            except AttributeError:
                warnings.warn(
                    "Prior specified only for Ainv. Inverting prior mean naively. "
                    "This operation is computationally costly! Specify an inverse "
                    "prior (mean) instead."
                )
                A0_mean = np.linalg.inv(Ainv0.mean)
            except NotImplementedError:
                A0_mean = linops.Identity(self.n)
                warnings.warn(
                    "Prior specified only for Ainv. Automatic prior mean inversion "
                    "not implemented, falling back to standard normal prior."
                )
            # Symmetric posterior correspondence
            A0_covfactor = self.A
            return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor

        # Prior on A specified
        elif A0 is not None and not isinstance(Ainv0, probnum.RandomVariable):
            if isinstance(A0, probnum.RandomVariable):
                A0_mean = A0.mean
                A0_covfactor = A0.cov.A
            else:
                self.is_calib_covclass = True
                A0_mean = A0
                A0_covfactor = A0  # Symmetric posterior correspondence
            try:
                if Ainv0 is not None:
                    Ainv0_mean = Ainv0
                elif isinstance(A0, probnum.RandomVariable):
                    Ainv0_mean = A0.mean.inv()
                else:
                    Ainv0_mean = A0.inv()
            except AttributeError:
                warnings.warn(
                    "Prior specified only for A. Inverting prior mean naively. "
                    "This operation is computationally costly! "
                    "Specify an inverse prior (mean)."
                )
                Ainv0_mean = np.linalg.inv(A0.mean)
            except NotImplementedError:
                Ainv0_mean = linops.Identity(self.n)
                warnings.warn(
                    "Prior specified only for A. Automatic prior mean inversion "
                    "failed, falling back to standard normal prior."
                )
            # Symmetric posterior correspondence
            Ainv0_covfactor = Ainv0_mean
            return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor
        # Both matrix priors on A and H specified via random variables
        elif isinstance(A0, probnum.RandomVariable) and isinstance(
            Ainv0, probnum.RandomVariable
        ):
            A0_mean = A0.mean
            A0_covfactor = A0.cov.A
            Ainv0_mean = Ainv0.mean
            Ainv0_covfactor = Ainv0.cov.A
            return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor
        else:
            raise NotImplementedError

    def _compute_trace_Ainv_covfactor0(self, Y, unc_scale):
        """
        Computes the trace of the prior covariance factor for the inverse view.

        Parameters
        ----------
        Y : np.ndarray, shape=(n,k)
            Observations.
        unc_scale : float
            Uncertainty scale :math:`\\psi` of the inverse view.

        Returns
        -------
        trace_Ainv_covfactor0 : float
            Trace of prior covariance factor.
        """
        # Initialization
        if Y is not None:
            k = Y.shape[1]
        else:
            k = 0
        if unc_scale is None:
            unc_scale = 0

        if isinstance(self.Ainv_covfactor0, linops.ScalarMult):
            # Scalar prior mean
            if self.is_calib_covclass and k > 0 and unc_scale != 0:
                _trace = self.Ainv_covfactor0.scalar * k
            else:
                _trace = self.Ainv_covfactor0.trace()
        else:
            # General prior mean
            if self.is_calib_covclass and k > 0 and unc_scale != 0:
                # General prior mean with calibration covariance class
                _trace = np.trace(
                    np.linalg.solve(
                        Y.T @ self.Ainv_mean0 @ Y,
                        Y.T @ self.Ainv_mean0 @ self.Ainv_mean0 @ Y,
                    )
                )
            else:
                _trace = self.Ainv_covfactor0.trace()
        if self.is_calib_covclass:
            # Additive term from uncertainty calibration
            _trace += unc_scale * (self.n - k)
        return _trace

    def _compute_trace_solution_covariance(self, bWb, Wb):
        """
        Computes the trace of the solution covariance
        :math:`\\tr(\\operatorname{Cov}[x])`

        Parameters
        ----------
        bWb : float
            Inner product of right hand side and the inverse covariance factor.
        Wb : np.ndarray
            Matrix-vector product between the inverse covariance factor and the right
            hand side.

        Returns
        -------
        trace_x_cov : float
            Trace of solution covariance.
        """
        # Trace of inverse covariance factor after k iterations
        return 0.5 * (bWb * self.trace_Ainv_covfactor + np.linalg.norm(Wb, ord=2) ** 2)

    def has_converged(self, iter, maxiter, resid=None, atol=None, rtol=None):
        """
        Check convergence of a linear solver.

        Evaluates a set of convergence criteria based on its input arguments to decide
        whether the iteration has converged.

        Parameters
        ----------
        iter : int
            Current iteration of solver.
        maxiter : int
            Maximum number of iterations
        resid : array-like
            Residual vector :math:`\\lVert r_i \\rVert = \\lVert Ax_i - b \\rVert` of
            the current iteration.
        atol : float
            Absolute residual tolerance. Stops if
            :math:`\\min(\\lVert r_i \\rVert, \\sqrt{\\operatorname{tr}(\\operatorname{Cov}(x))}) \\leq \\text{atol}`.
        rtol : float
            Relative residual tolerance. Stops if
            :math:`\\min(\\lVert r_i \\rVert, \\sqrt{\\operatorname{tr}(\\operatorname{Cov}(x))}) \\leq \\text{rtol} \\lVert b \\rVert`.

        Returns
        -------
        has_converged : bool
            True if the method has converged.
        convergence_criterion : str
            Convergence criterion which caused termination.
        """  # pylint: disable=line-too-long
        # maximum iterations
        if iter >= maxiter:
            warnings.warn(
                "Iteration terminated. Solver reached the maximum number of iterations."
            )
            return True, "maxiter"
        # residual below error tolerance
        resid_norm = np.linalg.norm(resid)
        b_norm = np.linalg.norm(self.b)
        if resid_norm <= atol:
            return True, "resid_atol"
        elif resid_norm <= rtol * b_norm:
            return True, "resid_rtol"
        # uncertainty-based
        if np.sqrt(self.trace_sol_cov) <= atol:
            return True, "tracecov_atol"
        elif np.sqrt(self.trace_sol_cov) <= rtol * b_norm:
            return True, "tracecov_rtol"
        else:
            return False, ""

    def _calibrate_uncertainty(self, S, sy, method):
        """
        Calibrate uncertainty based on the Rayleigh coefficients

        A regression model for the log-Rayleigh coefficient is built based on the
        collected observations. The degrees of freedom in the kernels of A and H are set
        according to the predicted log-Rayleigh coefficient for the remaining unexplored
        dimensions.

        Parameters
        ----------
        S : np.ndarray, shape=(n, k)
            Array of search directions
        sy : np.ndarray
            Array of inner products ``s_i'As_i``
        method : str
            Type of calibration method to use based on the Rayleigh quotient. Available
            calibration procedures are
            ====================================  ==================
             Most recent Rayleigh quotient         ``adhoc``
             Running (weighted) mean               ``weightedmean``
             GP regression for kernel matrices     ``gpkern``
            ====================================  ==================

        Returns
        -------
        phi : float
            Uncertainty scale of the null space of span(S) for the A view
        psi : float
            Uncertainty scale of the null space of span(Y) for the Ainv view
        """

        # Rayleigh quotient
        iters = np.arange(self.iter_ + 1)
        logR = np.log(sy) - np.log(np.einsum("nk,nk->k", S, S))

        # only calibrate if enough iterations for a regression model have been performed
        if self.iter_ > 1:
            if method == "adhoc":
                logR_pred = logR[-1]
            elif method == "weightedmean":
                deprecation_rate = 0.9
                logR_pred = logR * np.repeat(
                    deprecation_rate, self.iter_ + 1
                ) ** np.arange(self.iter_ + 1)
            elif method == "gpkern":
                # GP mean function via Weyl's result on spectra of Gram matrices for
                # differentiable kernels
                # ln(sigma(n)) ~= theta_0 - theta_1 ln(n)
                lnmap = GPy.core.Mapping(1, 1)
                lnmap.f = lambda n: np.log(n + 10 ** -16)
                lnmap.update_gradients = lambda a, b: None
                mf = GPy.mappings.Additive(
                    GPy.mappings.Constant(1, 1, value=0),
                    GPy.mappings.Compound(lnmap, GPy.mappings.Linear(1, 1)),
                )
                k = GPy.kern.RBF(input_dim=1, lengthscale=1, variance=1)
                m = GPy.models.GPRegression(
                    iters[:, None] + 1, logR[:, None], kernel=k, mean_function=mf
                )
                m.optimize(messages=False)

                # Predict Rayleigh quotient
                remaining_dims = np.arange(self.iter_, self.A.shape[0])[:, None]
                logR_pred = m.predict(remaining_dims + 1)[0].ravel()
            else:
                raise ValueError("Calibration method not recognized.")

            # Set uncertainty scale (degrees of freedom in calibration covariance class)
            Phi = (np.exp(np.mean(logR_pred))).item()
            Psi = (np.exp(-np.mean(logR_pred))).item()
        else:
            # For too few iterations take the most recent Rayleigh quotient
            Phi = np.exp(logR[-1])
            Psi = 1 / Phi

        return Phi, Psi

    def _get_calibration_covariance_update_terms(self, phi=None, psi=None):
        """
        For the calibration covariance class set the calibration update terms of the
        covariance in the null spaces of span(S) and span(Y) based on the degrees of
        freedom.
        """
        # Search directions and observations as arrays
        S = np.hstack(self.search_dir_list)
        Y = np.hstack(self.obs_list)

        def get_null_space_map(V, unc_scale):
            """
            Returns a function mapping to the null space of span(V), scaling with a
            single degree of freedom and mapping back.
            """

            def null_space_proj(x):
                try:
                    VVinvVx = np.linalg.solve(V.T @ V, V.T @ x)
                    return x - V @ VVinvVx
                except np.linalg.LinAlgError:
                    return np.zeros_like(x)

            # For a scalar uncertainty scale projecting to the null space twice is
            # equivalent to projecting once
            return lambda y: unc_scale * null_space_proj(y)

        # Compute calibration term in the A view as a linear operator with scaling from
        # degrees of freedom
        calibration_term_A = linops.LinearOperator(
            shape=(self.n, self.n), matvec=get_null_space_map(V=S, unc_scale=phi)
        )

        # Compute calibration term in the Ainv view as a linear operator with scaling
        # from degrees of freedom
        calibration_term_Ainv = linops.LinearOperator(
            shape=(self.n, self.n), matvec=get_null_space_map(V=Y, unc_scale=psi)
        )

        return calibration_term_A, calibration_term_Ainv

    def _get_output_randvars(self, Y_list, sy_list, phi=None, psi=None):
        """
        Return output random variables x, A, Ainv from their means and covariances.
        """

        if self.iter_ > 0:
            # Observations and inner products in A-space between actions
            Y = np.hstack(Y_list)
            sy = np.vstack(sy_list).ravel()

            # Posterior covariance factors
            if self.is_calib_covclass and (not phi is None) and (not psi is None):
                # Ensure prior covariance class only acts in span(S) like A
                def _matvec(x):
                    # First term of calibration covariance class: AS(S'AS)^{-1}S'A
                    return (Y * sy ** -1) @ (Y.T @ x.ravel())

                _A_covfactor0 = linops.LinearOperator(
                    shape=(self.n, self.n), matvec=_matvec
                )

                def _matvec(x):
                    # Term in covariance class: A_0^{-1}Y(Y'A_0^{-1}Y)^{-1}Y'A_0^{-1}
                    # TODO: for efficiency ensure that we dont have to compute
                    # (Y.T Y)^{-1} two times! For a scalar mean this is the same as in
                    # the null space projection
                    YAinv0Y_inv_YAinv0x = np.linalg.solve(
                        Y.T @ (self.Ainv_mean0 @ Y), Y.T @ (self.Ainv_mean0 @ x)
                    )
                    return self.Ainv_mean0 @ (Y @ YAinv0Y_inv_YAinv0x)

                _Ainv_covfactor0 = linops.LinearOperator(
                    shape=(self.n, self.n), matvec=_matvec
                )

                # Set degrees of freedom based on uncertainty calibration in unexplored
                # space
                (
                    calibration_term_A,
                    calibration_term_Ainv,
                ) = self._get_calibration_covariance_update_terms(phi=phi, psi=psi)

                _A_covfactor = (
                    _A_covfactor0 - self._A_covfactor_update_term + calibration_term_A
                )
                _Ainv_covfactor = (
                    _Ainv_covfactor0
                    - self._Ainv_covfactor_update_term
                    + calibration_term_Ainv
                )
            else:
                # No calibration
                _A_covfactor = self.A_covfactor
                _Ainv_covfactor = self.Ainv_covfactor
        else:
            # Converged before making any observations
            _A_covfactor = self.A_covfactor0
            _Ainv_covfactor = self.Ainv_covfactor0

        # Create output random variables
        A = rvs.Normal(mean=self.A_mean, cov=linops.SymmetricKronecker(A=_A_covfactor))

        Ainv = rvs.Normal(
            mean=self.Ainv_mean, cov=linops.SymmetricKronecker(A=_Ainv_covfactor)
        )
        # Induced distribution on x via Ainv
        # Exp(x) = Ainv b, Cov(x) = 1/2 (W b'Wb + Wbb'W)
        Wb = _Ainv_covfactor @ self.b
        bWb = np.squeeze(Wb.T @ self.b)

        def _mv(x):
            return 0.5 * (bWb * _Ainv_covfactor @ x + Wb @ (Wb.T @ x))

        cov_op = linops.LinearOperator(
            shape=(self.n, self.n), dtype=float, matvec=_mv, matmat=_mv
        )

        x = rvs.Normal(mean=self.x_mean.ravel(), cov=cov_op)

        # Compute trace of solution covariance: tr(Cov(x))
        self.trace_sol_cov = np.real_if_close(
            self._compute_trace_solution_covariance(bWb=bWb, Wb=Wb)
        ).item()

        return x, A, Ainv

    def _mean_update(self, u, v):
        """
        Linear operator implementing the symmetric rank 2 mean update (+= uv' + vu').
        """

        def mv(x):
            return u @ (v.T @ x) + v @ (u.T @ x)

        return linops.LinearOperator(shape=(self.n, self.n), matvec=mv, matmat=mv)

    def _covariance_update(self, u, Ws):
        """
        Linear operator implementing the symmetric rank 2 kernels update (-= Ws u^T).
        """

        def mv(x):
            return Ws @ (u.T @ x)

        return linops.LinearOperator(shape=(self.n, self.n), matvec=mv, matmat=mv)

    def solve(
        self, callback=None, maxiter=None, atol=None, rtol=None, calibration=None
    ):
        """
        Solve the linear system :math:`Ax=b`.

        Parameters
        ----------
        callback : function, optional
            User-supplied function called after each iteration of the linear solver. It
            is called as ``callback(xk, Ak, Ainvk, sk, yk, alphak, resid)`` and can be
            used to return quantities from the iteration. Note that depending on the
            function supplied, this can slow down the solver.
        maxiter : int
            Maximum number of iterations
        atol : float
            Absolute residual tolerance. Stops if
            :math:`\\min(\\lVert r_i \\rVert, \\sqrt{\\operatorname{tr}(\\operatorname{Cov}(x))}) \\leq \\text{atol}`.
        rtol : float
            Relative residual tolerance. Stops if
            :math:`\\min(\\lVert r_i \\rVert, \\sqrt{\\operatorname{tr}(\\operatorname{Cov}(x))}) \\leq \\text{rtol} \\lVert b \\rVert`.
        calibration : str or float, default=False
            If supplied calibrates the output via the given procedure or uncertainty
            scale. Available calibration procedures / choices are

            ====================================  ================
             No calibration                       None
             Provided scale                       float
             Most recent Rayleigh quotient        ``adhoc``
             Running (weighted) mean              ``weightedmean``
             GP regression for kernel matrices    ``gpkern``
            ====================================  ================

        Returns
        -------
        x : RandomVariable, shape=(n,) or (n, nrhs)
            Approximate solution :math:`x` to the linear system. Shape of the return
            matches the shape of ``b``.
        A : RandomVariable, shape=(n,n)
            Posterior belief over the linear operator.calibrate
        Ainv : RandomVariable, shape=(n,n)
            Posterior belief over the linear operator inverse :math:`H=A^{-1}`.
        info : dict
            Information on convergence of the solver.
        """  # pylint: disable=line-too-long
        # Initialization
        self.iter_ = 0
        resid = self.A @ self.x_mean - self.b

        # Initialize uncertainty calibration
        phi = None
        psi = None
        if calibration is None:
            pass
        elif (
            calibration is not None
            or calibration is not False
            and not self.is_calib_covclass
        ):
            warnings.warn(
                message="Cannot use calibration without a compatible covariance class."
            )
        elif isinstance(calibration, str) and self.is_calib_covclass:
            pass
        elif self.is_calib_covclass:
            if calibration < 0:
                raise ValueError("Calibration scale must be non-negative.")
            elif calibration == 0.0:
                pass
            else:
                phi = calibration
                psi = 1 / calibration

        # Trace of solution covariance
        _trace_Ainv_covfactor_update = 0
        self.trace_Ainv_covfactor = self._compute_trace_Ainv_covfactor0(
            Y=None, unc_scale=psi
        )

        # Create output random variables
        x, A, Ainv = self._get_output_randvars(
            Y_list=self.obs_list, sy_list=self.sy, phi=phi, psi=psi
        )

        # Iteration with stopping criteria
        while True:
            # Check convergence
            _has_converged, _conv_crit = self.has_converged(
                iter=self.iter_, maxiter=maxiter, resid=resid, atol=atol, rtol=rtol
            )
            if _has_converged:
                break

            # Compute search direction (with implicit reorthogonalization) via policy
            search_dir = -self.Ainv_mean @ resid
            self.search_dir_list.append(search_dir)

            # Perform action and observe
            obs = self.A @ search_dir
            self.obs_list.append(obs)

            # Compute step size
            sy = search_dir.T @ obs
            step_size = -np.squeeze((search_dir.T @ resid) / sy)
            self.sy.append(sy)

            # Step and residual update
            self.x_mean = self.x_mean + step_size * search_dir
            resid = resid + step_size * obs

            # (Symmetric) mean and covariance updates
            Vs = self.A_covfactor @ search_dir
            delta_A = obs - self.A_mean @ search_dir
            u_A = Vs / (search_dir.T @ Vs)
            v_A = delta_A - 0.5 * (search_dir.T @ delta_A) * u_A

            Wy = self.Ainv_covfactor @ obs
            delta_Ainv = search_dir - self.Ainv_mean @ obs
            yWy = np.squeeze(obs.T @ Wy)
            u_Ainv = Wy / yWy
            v_Ainv = delta_Ainv - 0.5 * (obs.T @ delta_Ainv) * u_Ainv

            # Rank 2 mean updates (+= uv' + vu')
            self.A_mean = linops.aslinop(self.A_mean) + self._mean_update(u=u_A, v=v_A)
            self.Ainv_mean = linops.aslinop(self.Ainv_mean) + self._mean_update(
                u=u_Ainv, v=v_Ainv
            )

            # Rank 1 covariance Kronecker factor update (-= u_A(Vs)' and -= u_Ainv(Wy)')
            if self.iter_ == 0:
                self._A_covfactor_update_term = self._covariance_update(u=u_A, Ws=Vs)
                self._Ainv_covfactor_update_term = self._covariance_update(
                    u=u_Ainv, Ws=Wy
                )
            else:
                self._A_covfactor_update_term = (
                    self._A_covfactor_update_term
                    + self._covariance_update(u=u_A, Ws=Vs)
                )
                self._Ainv_covfactor_update_term = (
                    self._Ainv_covfactor_update_term
                    + self._covariance_update(u=u_Ainv, Ws=Wy)
                )
            self.A_covfactor = (
                linops.aslinop(self.A_covfactor0) - self._A_covfactor_update_term
            )
            self.Ainv_covfactor = (
                linops.aslinop(self.Ainv_covfactor0) - self._Ainv_covfactor_update_term
            )

            # Calibrate uncertainty based on Rayleigh quotient
            if isinstance(calibration, str) and self.is_calib_covclass:
                phi, psi = self._calibrate_uncertainty(
                    S=np.hstack(self.search_dir_list),
                    sy=np.vstack(self.sy).ravel(),
                    method=calibration,
                )

            # Update trace of solution covariance: tr(Cov(Hb))
            _trace_Ainv_covfactor_update += 1 / yWy * np.squeeze(Wy.T @ Wy)
            self.trace_Ainv_covfactor = np.real_if_close(
                self._compute_trace_Ainv_covfactor0(
                    Y=np.hstack(self.obs_list), unc_scale=psi
                )
                - _trace_Ainv_covfactor_update
            ).item()

            # Create output random variables
            x, A, Ainv = self._get_output_randvars(
                Y_list=self.obs_list, sy_list=self.sy, phi=phi, psi=psi
            )

            # Callback function used to extract quantities from iteration
            if callback is not None:
                callback(
                    xk=x,
                    Ak=A,
                    Ainvk=Ainv,
                    sk=search_dir,
                    yk=obs,
                    alphak=step_size,
                    resid=resid,
                )

            # Iteration increment
            self.iter_ += 1

        # Log information on solution
        info = {
            "iter": self.iter_,
            "maxiter": maxiter,
            "resid_l2norm": np.linalg.norm(resid, ord=2),
            "trace_sol_cov": self.trace_sol_cov,
            "conv_crit": _conv_crit,
            "rel_cond": None,  # TODO: matrix condition from solver (see scipy solvers)
        }

        return x, A, Ainv, info


class NoisySymmetricMatrixBasedSolver(MatrixBasedSolver):
    """
    Solver iteration of the noisy symmetric probabilistic linear solver.

    Implements the solve iteration of the symmetric matrix-based probabilistic linear
    solver taking into account noisy matrix-vector products :math:`y_k = (A + E_k)s_k`
    as described in [1]_ and [2]_.

    Parameters
    ----------
    A : LinearOperator or RandomVariable, shape=(n,n)
        The square matrix or linear operator of the linear system.
    b : array_like, shape=(n,) or (n, nrhs)
        Right-hand side vector or matrix in :math:`A x = b`.
    A0 : array-like or LinearOperator or RandomVariable, shape=(n, n), optional
        A square matrix, linear operator or random variable representing the prior
        belief over the linear operator :math:`A`. If an array or linear operator is
        given, a prior distribution is chosen automatically.
    Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
        A square matrix, linear operator or random variable representing the prior
        belief over the inverse :math:`H=A^{-1}`. This can be viewed as taking the form
        of a pre-conditioner. If an array or linear operator is given, a prior
        distribution is chosen automatically.
    x0 : array-like, or RandomVariable, shape=(n,) or (n, nrhs)
        Optional. Prior belief for the solution of the linear system. Will be ignored if
        ``Ainv0`` is given.

    Returns
    -------
    A : RandomVariable
        Posterior belief over the linear operator.
    Ainv : RandomVariable
        Posterior belief over the inverse linear operator.
    x : RandomVariable
        Posterior belief over the solution of the linear system.
    info : dict
        Information about convergence and the solution.

    References
    ----------
    .. [1] Wenger, J., de Roos, F. and Hennig, P., Probabilistic Solution of Noisy
       Linear Systems, 2020
    .. [2] Hennig, P., Probabilistic Interpretation of Linear Solvers, *SIAM Journal on
       Optimization*, 2015, 25, 234-260

    See Also
    --------
    SymmetricMatrixBasedSolver :
        Class implementing the symmetric probabilistic linear solver.
    """

    def __init__(self, A, b, A0=None, Ainv0=None, x0=None):

        # Transform right hand side to random variable
        if not isinstance(b, probnum.RandomVariable):
            _b = probnum.asrandvar(b)
        else:
            _b = b

        super().__init__(A=A, b=_b, x0=x0)

        # Get or initialize prior parameters
        (
            A0_mean,
            A0_covfactor,
            Ainv0_mean,
            Ainv0_covfactor,
            b_mean,
        ) = self._get_prior_params(A0=A0, Ainv0=Ainv0, x0=x0, b=_b)

        # Matrix prior parameters
        self.A0_mean = linops.aslinop(A0_mean)
        self.A_mean = linops.aslinop(A0_mean)
        self.A0_covfactor = A0_covfactor
        self.Ainv0_mean = linops.aslinop(Ainv0_mean)
        self.Ainv_mean = linops.aslinop(Ainv0_mean)
        self.Ainv0_covfactor = Ainv0_covfactor
        self.b_mean = b_mean

        # Induced distribution on x via Ainv
        # Exp = x = A^-1 b, Cov = 1/2 (W b'Wb + Wbb'W)
        Wb = Ainv0_covfactor @ self.b_mean
        bWb = np.squeeze(Wb.T @ self.b_mean)

        def _mv(x):
            return 0.5 * (bWb * Ainv0_covfactor @ x + Wb @ (Wb.T @ x))

        self.x_cov = linops.LinearOperator(
            shape=(self.n, self.n), dtype=float, matvec=_mv, matmat=_mv
        )
        if isinstance(x0, np.ndarray):
            self.x_mean = x0
        elif x0 is None:
            self.x_mean = Ainv0_mean @ self.b_mean
        else:
            raise NotImplementedError
        self.x0 = self.x_mean

    def _get_prior_params(self, A0, Ainv0, x0, b):
        """
        Get the parameters of the matrix priors on A and H.

        Retrieves and / or initializes prior parameters of ``A0`` and ``Ainv0``.

        Parameters
        ----------
        A0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
            A square matrix, linear operator or random variable representing the prior
            belief over the linear operator :math:`A`. If an array or linear operator is
            given, a prior distribution is chosen automatically.
        Ainv0 : array-like or LinearOperator or RandomVariable, shape=(n,n), optional
            A square matrix, linear operator or random variable representing the prior
            belief over the inverse :math:`H=A^{-1}`. This can be viewed as taking the
            form of a pre-conditioner. If an array or linear operator is given, a prior
            distribution is chosen automatically.
        x0 : array-like, or RandomVariable, shape=(n,)
            Optional. Prior belief for the solution of the linear system. Will be
            ignored if ``A0`` or ``Ainv0`` is given.
        b : RandomVariable, shape=(n,) or (n, nrhs)
            Right-hand side random variable `b` in :math:`A x = b`.

        Returns
        -------
        A0_mean : array-like or LinearOperator, shape=(n,n)
            Prior mean of the linear operator :math:`A`.
        A0_covfactor : array-like or LinearOperator, shape=(n,n)
            Factor :math:`W^A` of the symmetric Kronecker product prior covariance
            :math:`W^A \\otimes_s W^A` of :math:`A`.
        Ainv0_mean : array-like or LinearOperator, shape=(n,n)
            Prior mean of the linear operator :math:`H`.
        Ainv0_covfactor : array-like or LinearOperator, shape=(n,n)
            Factor :math:`W^H` of the symmetric Kronecker product prior covariance
            :math:`W^H \\otimes_s W^H` of :math:`H`.
        b_mean : array-like, shape=(n,nrhs)
            Prior mean of the right hand side :math:`b`.
        """

        # Right hand side mean
        b_mean = b.sample(1)  # TODO: build prior model for rhs and change to b.mean

        # No matrix priors specified
        if A0 is None and Ainv0 is None:
            # No prior information given
            if x0 is None:
                Ainv0_mean = linops.Identity(shape=self.n)
                Ainv0_covfactor = linops.Identity(shape=self.n)
                # Standard normal covariance
                A0_mean = linops.Identity(shape=self.n)
                A0_covfactor = linops.Identity(shape=self.n)
                # TODO: should this be a sample from A to achieve symm. posterior
                # correspondence?
                return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor, b_mean
            # Construct matrix priors from initial guess x0
            elif isinstance(x0, np.ndarray):
                # Sample from linear operator for prior construction
                if isinstance(self.A, probnum.RandomVariable):
                    _A = self.A.sample([1])[0]
                else:
                    _A = self.A
                A0_mean, Ainv0_mean = self._construct_symmetric_matrix_prior_means(
                    A=_A, x0=x0, b=b_mean
                )
                Ainv0_covfactor = Ainv0_mean
                # Standard normal covariance
                A0_covfactor = linops.Identity(shape=self.n)
                # TODO: should this be a sample from A to achieve symm. posterior
                # correspondence?
                return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor, b_mean
            elif isinstance(x0, probnum.RandomVariable):
                raise NotImplementedError

        # Prior on Ainv specified
        if not isinstance(A0, probnum.RandomVariable) and Ainv0 is not None:
            if isinstance(Ainv0, probnum.RandomVariable):
                Ainv0_mean = Ainv0.mean
                Ainv0_covfactor = Ainv0.cov.A
            else:
                Ainv0_mean = Ainv0
                Ainv0_covfactor = Ainv0  # Symmetric posterior correspondence
            try:
                if A0 is not None:
                    A0_mean = A0
                elif isinstance(Ainv0, probnum.RandomVariable):
                    A0_mean = Ainv0.mean.inv()
                else:
                    A0_mean = Ainv0.inv()
            except AttributeError:
                warnings.warn(
                    "Prior specified only for Ainv. Inverting prior mean naively. "
                    "This operation is computationally costly! Specify an inverse "
                    "prior (mean) instead."
                )
                A0_mean = np.linalg.inv(Ainv0.mean)
            except NotImplementedError:
                A0_mean = linops.Identity(self.n)
                warnings.warn(
                    "Prior specified only for Ainv. Automatic prior mean inversion "
                    "not implemented, falling back to standard normal prior."
                )
            # Standard normal covariance
            A0_covfactor = linops.Identity(shape=self.n)
            # TODO: should this be a sample from A to achieve symm. posterior
            # correspondence?
            return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor, b_mean

        # Prior on A specified
        elif A0 is not None and not isinstance(Ainv0, probnum.RandomVariable):
            if isinstance(A0, probnum.RandomVariable):
                A0_mean = A0.mean
                A0_covfactor = A0.cov.A
            else:
                A0_mean = A0
                A0_covfactor = A0  # Symmetric posterior correspondence
            try:
                if Ainv0 is not None:
                    Ainv0_mean = Ainv0
                elif isinstance(A0, probnum.RandomVariable):
                    Ainv0_mean = A0.mean.inv()
                else:
                    Ainv0_mean = A0.inv()
            except AttributeError:
                warnings.warn(
                    "Prior specified only for A. Inverting prior mean naively. "
                    "This operation is computationally costly! Specify an inverse "
                    "prior (mean) instead."
                )
                Ainv0_mean = np.linalg.inv(A0.mean)
            except NotImplementedError:
                Ainv0_mean = linops.Identity(self.n)
                warnings.warn(
                    "Prior specified only for A. Automatic prior mean inversion "
                    "failed, falling back to standard normal prior."
                )
            # Symmetric posterior correspondence
            Ainv0_covfactor = Ainv0_mean
            return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor, b_mean
        # Both matrix priors on A and H specified via random variables
        elif isinstance(A0, probnum.RandomVariable) and isinstance(
            Ainv0, probnum.RandomVariable
        ):
            A0_mean = A0.mean
            A0_covfactor = A0.cov.A
            Ainv0_mean = Ainv0.mean
            Ainv0_covfactor = Ainv0.cov.A
            return A0_mean, A0_covfactor, Ainv0_mean, Ainv0_covfactor, b_mean
        else:
            raise NotImplementedError

    def has_converged(self, iter, maxiter, atol=None, rtol=None):
        """
        Check convergence of a linear solver.

        Evaluates a set of convergence criteria based on its input arguments to decide
        whether the iteration has converged.

        Parameters
        ----------
        iter : int
            Current iteration of solver.
        maxiter : int
            Maximum number of iterations
        atol : float
            Absolute tolerance for the uncertainty about the solution estimate. Stops if
            :math:`\\sqrt{\\text{tr}(\\Sigma)}  \\leq \\text{atol}`, where
            :math:`\\Sigma` is the covariance of the solution :math:`x`.
        rtol : float
            Relative tolerance for the uncertainty about the solution estimate. Stops if
            :math:`\\sqrt{\\text{tr}(\\Sigma)} \\leq \\text{rtol} \\lVert x_i \\rVert`,
            where :math:`\\Sigma` is the covariance of the solution :math`x` and
            :math:`x_i` its mean.

        Returns
        -------
        has_converged : bool
            True if the method has converged.
        convergence_criterion : str
            Convergence criterion which caused termination.
        """
        # maximum iterations
        if iter >= maxiter:
            warnings.warn(
                "Iteration terminated. Solver reached the maximum number of iterations."
            )
            return True, "maxiter"
        # uncertainty-based
        if isinstance(self.x_cov, linops.LinearOperator):
            sqrttracecov = np.sqrt(self.x_cov.trace())
        else:
            sqrttracecov = np.sqrt(np.trace(self.x_cov))
        if sqrttracecov <= atol:
            return True, "covar_atol"
        elif sqrttracecov <= rtol * np.linalg.norm(self.x_mean):
            return True, "covar_rtol"
        else:
            return False, ""

    def solve(
        self,
        callback=None,
        maxiter=None,
        atol=10 ** -6,
        rtol=10 ** -6,
        noise_scale=None,
        **kwargs
    ):
        """
        Solve the linear system :math:`Ax=b`.

        Parameters
        ----------
        callback : function, optional
            User-supplied function called after each iteration of the linear solver. It
            is called as ``callback(xk, Ak, Ainvk, sk, yk, alphak, resid, noise_scale)``
            and can be used to return quantities from the iteration. Note that depending
            on the function supplied, this can slow down the solver.
        maxiter : int
            Maximum number of iterations
        atol : float
            Absolute tolerance for the uncertainty about the solution estimate. Stops if
            :math:`\\sqrt{\\text{tr}(\\Sigma)}  \\leq \\text{atol}`, where
            :math:`\\Sigma` is the covariance of the solution :math:`x`.
        rtol : float
            Relative tolerance for the uncertainty about the solution estimate. Stops if
            :math:`\\sqrt{\\text{tr}(\\Sigma)} \\leq \\text{rtol} \\lVert x_i \\rVert`,
            where :math:`\\Sigma` is the covariance of the solution :math`x` and
            :math:`x_i` its mean.
        noise_scale : float
            Assumed (initial) noise scale :math:`\\varepsilon^2`.

        Returns
        -------
        x : RandomVariable, shape=(n,) or (n, nrhs)
            Approximate solution :math:`x` to the linear system. Shape of the return
            matches the shape of ``b``.
        A : RandomVariable, shape=(n,n)
            Posterior belief over the linear operator.
        Ainv : RandomVariable, shape=(n,n)
            Posterior belief over the linear operator inverse :math:`H=A^{-1}`.
        info : dict
            Information on convergence of the solver.
        """
        raise NotImplementedError



"""Abstract Base Class for posteriors over states after applying filtering/smoothing"""
from abc import ABC, abstractmethod


class FiltSmoothPosterior(ABC):
    """Posterior Distribution over States after Filtering/Smoothing"""

    @abstractmethod
    def __call__(self, location):
        """Evaluate the time-continuous posterior for a given location

        Parameters
        ----------
        location : float
            Location, or time, at which to evaluate the posterior.

        Returns
        -------
        rv : `RandomVariable`
        """
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        """Length of the discrete-time solution

        Corresponds to the number of filtering/smoothing steps
        """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, idx):
        """Return the corresponding index/slice of the discrete-time solution"""
        raise NotImplementedError



"""
Gaussian filtering and smoothing based on making intractable quantities
tractable through Taylor-method approximations, e.g. linearization.
"""
import numpy as np

from probnum.filtsmooth.gaussfiltsmooth.gaussfiltsmooth import (
    GaussFiltSmooth,
    linear_discrete_update,
)
from probnum.random_variables import Normal
from probnum.filtsmooth.statespace import LinearSDEModel, DiscreteGaussianModel

from probnum.filtsmooth.gaussfiltsmooth._utils import is_cont_disc, is_disc_disc


class ExtendedKalman(GaussFiltSmooth):
    """
    Factory method for Kalman filters.
    """

    def __new__(cls, dynamod, measmod, initrv, **kwargs):

        if cls is ExtendedKalman:
            if is_cont_disc(dynamod, measmod):
                return _ContDiscExtendedKalman(dynamod, measmod, initrv, **kwargs)
            if is_disc_disc(dynamod, measmod):
                return _DiscDiscExtendedKalman(dynamod, measmod, initrv, **kwargs)
            else:
                errmsg = (
                    "Cannot instantiate Extended Kalman filter with "
                    "given dynamic model and measurement model."
                )
                raise ValueError(errmsg)
        else:
            return super().__new__(cls)


class _ContDiscExtendedKalman(ExtendedKalman):
    """
    Continuous-discrete extended Kalman filtering and smoothing.
    """

    def __init__(self, dynamod, measmod, initrv, **kwargs):
        if not issubclass(type(dynamod), LinearSDEModel):
            raise ValueError(
                "This implementation of ContDiscExtendedKalman "
                "requires a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianModel):
            raise ValueError(
                "ContDiscExtendedKalman requires a Gaussian measurement model."
            )
        if "cke_nsteps" in kwargs.keys():
            self.cke_nsteps = kwargs["cke_nsteps"]
        else:
            self.cke_nsteps = 1
        super().__init__(dynamod, measmod, initrv)

    def predict(self, start, stop, randvar, **kwargs):
        step = (stop - start) / self.cke_nsteps
        return self.dynamicmodel.chapmankolmogorov(start, stop, step, randvar, **kwargs)

    def update(self, time, randvar, data, **kwargs):
        return _discrete_extkalman_update(
            time, randvar, data, self.measurementmodel, **kwargs
        )


class _DiscDiscExtendedKalman(ExtendedKalman):
    def __init__(self, dynamod, measmod, initrv, **kwargs):
        """
        Checks that dynamod and measmod are linear and moves on.
        """
        if not issubclass(type(dynamod), DiscreteGaussianModel):
            raise ValueError(
                "DiscDiscExtendedKalmanFilter requires " "a linear dynamic model."
            )
        if not issubclass(type(measmod), DiscreteGaussianModel):
            raise ValueError(
                "DiscDiscExtendedKalmanFilter requires " "a linear measurement model."
            )
        super().__init__(dynamod, measmod, initrv)

    def predict(self, start, stop, randvar, **kwargs):
        mean, covar = randvar.mean, randvar.cov
        if np.isscalar(mean) and np.isscalar(covar):
            mean, covar = mean * np.ones(1), covar * np.eye(1)
        diffmat = self.dynamod.diffusionmatrix(start, **kwargs)
        jacob = self.dynamod.jacobian(start, mean, **kwargs)
        mpred = self.dynamod.dynamics(start, mean, **kwargs)
        crosscov = covar @ jacob.T
        cpred = jacob @ crosscov + diffmat
        return Normal(mpred, cpred), crosscov

    def update(self, time, randvar, data, **kwargs):
        return _discrete_extkalman_update(
            time, randvar, data, self.measurementmodel, **kwargs
        )


def _discrete_extkalman_update(time, randvar, data, measmod, **kwargs):
    mpred, cpred = randvar.mean, randvar.cov
    if np.isscalar(mpred) and np.isscalar(cpred):
        mpred, cpred = mpred * np.ones(1), cpred * np.eye(1)
    jacob = measmod.jacobian(time, mpred, **kwargs)
    meascov = measmod.diffusionmatrix(time, **kwargs)
    meanest = measmod.dynamics(time, mpred, **kwargs)
    return linear_discrete_update(meanest, cpred, data, meascov, jacob, mpred)



import numpy as np

from probnum.random_variables import Normal
from probnum.diffeq import odesolver
from probnum.diffeq.odefiltsmooth.prior import ODEPrior
from probnum.diffeq.odesolution import ODESolution


class GaussianIVPFilter(odesolver.ODESolver):
    """
    ODE solver that behaves like a Gaussian filter.


    This is based on continuous-discrete Gaussian filtering.

    Note: this is specific for IVPs and does not apply without
    further considerations to, e.g., BVPs.
    """

    def __init__(self, ivp, gaussfilt, with_smoothing):
        """
        gaussfilt : gaussianfilter.GaussianFilter object,
            e.g. the return value of ivp_to_ukf(), ivp_to_ekf1().

        Notes
        -----
        * gaussfilt.dynamicmodel contains the prior,
        * gaussfilt.measurementmodel contains the information about the
        ODE right hand side function,
        * gaussfilt.initialdistribution contains the information about
        the initial values.
        """
        if not issubclass(type(gaussfilt.dynamicmodel), ODEPrior):
            raise ValueError("Please initialise a Gaussian filter with an ODEPrior")
        self.gfilt = gaussfilt
        self.sigma_squared_global = 0.0
        self.sigma_squared_current = 0.0
        self.with_smoothing = with_smoothing
        super().__init__(ivp)

    def initialise(self):
        return self.ivp.t0, self.gfilt.initialrandomvariable

    def step(self, t, t_new, current_rv, **kwargs):
        """Gaussian IVP filter step as nonlinear Kalman filtering with zero data."""
        pred_rv, _ = self.gfilt.predict(t, t_new, current_rv, **kwargs)
        zero_data = 0.0
        filt_rv, meas_cov, crosscov, meas_mean = self.gfilt.update(
            t_new, pred_rv, zero_data, **kwargs
        )
        errorest, self.sigma_squared_current = self._estimate_error(
            filt_rv.mean, crosscov, meas_cov, meas_mean
        )
        return filt_rv, errorest

    def method_callback(self, time, current_guess, current_error):
        """Update the sigma-squared (ssq) estimate."""
        self.sigma_squared_global = (
            self.sigma_squared_global
            + (self.sigma_squared_current - self.sigma_squared_global) / self.num_steps
        )

    def postprocess(self, times, rvs):
        """
        Rescale covariances with sigma square estimate,
        (if specified) smooth the estimate, return ODESolution.
        """
        rvs = self._rescale(rvs)
        odesol = super().postprocess(times, rvs)
        if self.with_smoothing is True:
            odesol = self._odesmooth(ode_solution=odesol)
        return odesol

    def _rescale(self, rvs):
        """Rescales covariances according to estimate sigma squared value."""
        rvs = [Normal(rv.mean, self.sigma_squared_global * rv.cov) for rv in rvs]
        return rvs

    def _odesmooth(self, ode_solution, **kwargs):
        """
        Smooth out the ODE-Filter output.

        Be careful about the preconditioning: the GaussFiltSmooth object
        only knows the state space with changed coordinates!

        Parameters
        ----------
        filter_solution: ODESolution

        Returns
        -------
        smoothed_solution: ODESolution
        """
        ivp_filter_posterior = ode_solution._kalman_posterior
        ivp_smoother_posterior = self.gfilt.smooth(ivp_filter_posterior, **kwargs)

        smoothed_solution = ODESolution(
            times=ivp_smoother_posterior.locations,
            rvs=ivp_smoother_posterior.state_rvs,
            solver=ode_solution._solver,
        )

        return smoothed_solution

    def undo_preconditioning(self, rv):
        ipre = self.gfilt.dynamicmodel.invprecond
        newmean = ipre @ rv.mean
        newcov = ipre @ rv.cov @ ipre.T
        newrv = Normal(newmean, newcov)
        return newrv

    def _estimate_error(self, currmn, ccest, covest, mnest):
        """
        Error estimate.

        Estimates error as whitened residual using the
        residual as weight vector.

        THIS IS NOT A PERFECT ERROR ESTIMATE, but this is a question of
        research, not a question of implementation as of now.
        """
        std_like = np.linalg.cholesky(covest)
        whitened_res = np.linalg.solve(std_like, mnest)
        ssq = whitened_res @ whitened_res / mnest.size
        abserrors = np.abs(whitened_res)
        errorest = self._rel_and_abs_error(abserrors, currmn)
        return errorest, ssq

    def _rel_and_abs_error(self, abserrors, currmn):
        """
        Returns maximum of absolute and relative error.
        This way, both are guaranteed to be satisfied.
        """
        ordint, spatialdim = self.gfilt.dynamicmodel.ordint, self.ivp.ndim
        h0_1d = np.eye(ordint + 1)[:, 0].reshape((1, ordint + 1))
        projmat = np.kron(np.eye(spatialdim), h0_1d)
        weights = np.ones(len(abserrors))
        rel_error = (
            (abserrors / np.abs(projmat @ currmn)) @ weights / np.linalg.norm(weights)
        )
        abs_error = abserrors @ weights / np.linalg.norm(weights)
        return np.maximum(rel_error, abs_error)

    @property
    def prior(self):
        return self.gfilt.dynamicmodel



"""
Ordinary differential equations.

Subclassed by the types of ODEs: IVPs, BVPs and whatever you
can imagine.

See Also
--------
IVP : Initial value problems (ivp.py)
"""

from abc import ABC, abstractmethod


class ODE(ABC):
    """
    Ordinary differential equations.

    Extended by the types of ODEs, e.g. IVPs, BVPs.
    This class describes systems of irst order ordinary differential
    equations (ODEs),

    .. math:: \\dot y(t) = f(t, y(t)), \\quad t \\in [t_0, T].

    It provides options for defining custom right-hand side (RHS)
    functions, their Jacobians and closed form solutions.

    Parameters
    ----------
    timespan : (float, float)
        Time span of IVP.
    rhs : callable, signature: ``(t, y, **kwargs)``
        RHS function
        :math:`f : [t_0, T] \times \\mathbb{R}^d \\rightarrow \\mathbb{R}^d`
        of the ODE system. As such it takes a float and an
        np.ndarray of shape (d,) and returns a np.ndarray
        of shape (d,). As of now, no vectorization is supported
        (nor needed).
    jac : callable, signature: ``(t, y, **kwargs)``, optional
        Jacobian of RHS function
        :math:`J_f : [0, T] \\times \\mathbb{R}^d \\rightarrow \\mathbb{R}^d`
        of the ODE system. As such it takes a float and an
        np.ndarray of shape (d,) and returns a np.ndarray
        of shape (d,). As of now, no vectorization is supported
        (nor needed).
    sol : callable, signature: ``(t, **kwargs)``, optional
        Solution of the ODE system. Only well-defined in subclasses like
        IVP or BVP.

    See Also
    --------
    IVP : Extends ODE for initial value problems.
    """

    def __init__(self, timespan, rhs, jac=None, hess=None, sol=None):
        """
        Initialises basic ODE attributes.

        Essentially, all but the initial/boundary distribution
        which then determine the type of problem.

        We take timespan as a tuple to mimic the
        scipy.solve_ivp interface.
        """
        self._t0, self._tmax = timespan
        self._rhs = rhs
        self._jac = jac
        self._hess = hess
        self._sol = sol

    def __call__(self, t, y, **kwargs):
        """
        Piggybacks on self.rhs(t, y).
        """
        return self.rhs(t, y, **kwargs)

    def rhs(self, t, y, **kwargs):
        """
        Evaluates model function f.
        """
        return self._rhs(t, y, **kwargs)

    def jacobian(self, t, y, **kwargs):
        """
        Jacobian of model function f.
        """
        if self._jac is None:
            raise NotImplementedError
        else:
            return self._jac(t, y, **kwargs)

    def hessian(self, t, y, **kwargs):
        """
        Hessian of model function f.

        For :math:`d=3`, the Hessian
        :math:`H_f(t, y) \\in \\mathbb{R}^{3 \\times 3 \\times 3}`
        is expected be evaluated as

        .. math:: H_f(t, y) = \\left[H_{f_1}(t, y), H_{f_2}(t, y), H_{f_3}(t, y) \\right]^\\top

        since for any directions :math:`v_1, v_2` the outcome of
        :math:`H_f(t_0, y_0) \\cdot v_1 \\cdot v_2` is expected to contain
        the incline of :math:`f_i` in direction :math:`(v_1, v_2)`.
        """  # pylint: disable=line-too-long
        if self._hess is None:
            raise NotImplementedError
        else:
            return self._hess(t, y, **kwargs)

    def solution(self, t, **kwargs):
        """
        Solution of the IVP.
        """
        if self._sol is None:
            raise NotImplementedError
        else:
            return self._sol(t, **kwargs)

    @property
    def t0(self):
        return self._t0

    @property
    def tmax(self):
        return self._tmax

    @property
    def timespan(self):
        """
        Returns :math:`(t_0, T)` as ``[self.t0, self.tmax]``.
        Mainly here to provide an interface to scipy.integrate.
        Both :math:`t_0` and :math:`T` can be accessed via
        ``self.t0`` and ``self.tmax`` respectively.
        """
        return [self._t0, self._tmax]

    @property
    @abstractmethod
    def ndim(self):
        """
        Abstract, in order to force subclassing.
        """
        raise NotImplementedError



"""
Convenience functions for Gaussian filtering and smoothing.

We support the following methods:
    - ekf0: Extended Kalman filtering based on a zero-th order Taylor
        approximation [1]_, [2]_, [3]_. Also known as "PFOS".
    - ekf1: Extended Kalman filtering [3]_.
    - ukf: Unscented Kalman filtering [3]_.
    - eks0: Extended Kalman smoothing based on a zero-th order Taylor
        approximation [4]_.
    - eks1: Extended Kalman smoothing [4]_.
    - uks: Unscented Kalman smoothing.

References
----------
.. [1] https://arxiv.org/pdf/1610.05261.pdf
.. [2] https://arxiv.org/abs/1807.09737
.. [3] https://arxiv.org/abs/1810.03440
.. [4] https://arxiv.org/pdf/2004.00623.pdf
"""

from probnum.diffeq import steprule
from probnum.diffeq.odefiltsmooth import prior, ivp2filter
from probnum.diffeq.odefiltsmooth.ivpfiltsmooth import GaussianIVPFilter


def probsolve_ivp(
    ivp,
    method="ekf0",
    which_prior="ibm1",
    tol=None,
    step=None,
    firststep=None,
    precond_step=None,
    **kwargs
):
    """
    Solve initial value problem with Gaussian filtering and smoothing.

    Numerically computes a Gauss-Markov process which solves numerically
    the initial value problem (IVP) based on a system of first order
    ordinary differential equations (ODEs)

    .. math:: \\dot x(t) = f(t, x(t)), \\quad x(t_0) = x_0,
        \\quad t \\in [t_0, T]

    by regarding it as a (nonlinear) Gaussian filtering (and smoothing)
    problem [3]_. For some configurations it recovers certain multistep
    methods [1]_.
    Convergence rates of filtering [2]_ and smoothing [4]_ are
    comparable to those of methods of Runge-Kutta type.


    This function turns a prior-string into an :class:`ODEPrior`, a
    method-string into a filter/smoother of class :class:`GaussFiltSmooth`, creates a
    :class:`GaussianIVPFilter` object and calls the :meth:`solve()` method. For
    advanced usage we recommend to do this process manually which
    enables advanced methods of tuning the algorithm.

    This function supports the methods:
    extended Kalman filtering based on a zero-th order Taylor
    approximation (EKF0),
    extended Kalman filtering (EKF1),
    unscented Kalman filtering (UKF),
    extended Kalman smoothing based on a zero-th order Taylor
    approximation (EKS0),
    extended Kalman smoothing (EKS1), and
    unscented Kalman smoothing (UKS).

    Arguments
    ---------
    ivp : IVP
        Initial value problem to be solved.
    step : float
        Step size :math:`h` of the solver. This defines the
        discretisation mesh as each proposed step is equal to :math:`h`
        and all proposed steps are accepted.
        Only one of out of ``step`` and ``tol`` is set.
    tol : float
        Tolerance :math:`\\varepsilon` of the adaptive step scheme.
        We implement the scheme proposed by Schober et al., accepting a
        step if the absolute as well as the relative error estimate are
        smaller than the tolerance,
        :math:`\\max\\{e, e / |y|\\} \\leq \\varepsilon`.
        Only one of out of ``step`` and ``tol`` is set.
    which_prior : str, optional
        Which prior is to be used. Default is an IBM(1), further support
        for IBM(:math:`q`), IOUP(:math:`q`), Matern(:math:`q+1/2`),
        :math:`q\\in\\{1, 2, 3, 4\\}` is provided. The available
        options are

        ======================  ========================================
         IBM(:math:`q`)         ``'ibm1'``, ``'ibm2'``, ``'ibm3'``,
                                ``'ibm4'``
         IOUP(:math:`q`)        ``'ioup1'``, ``'ioup2'``, ``'ioup3'``,
                                ``'ioup4'``
         Matern(:math:`q+0.5`)  ``'matern32'``, ``'matern52'``,
                                ``'matern72'``, ``'matern92'``
        ======================  ========================================

        The type of prior relates to prior assumptions about the
        derivative of the solution. The IBM(:math:`q`) prior leads to a
        :math:`q`-th order method that is recommended if little to no
        prior information about the solution is available. On the other
        hand, if the :math:`q`-th derivative is expected to regress to
        zero, an IOUP(:math:`q`) prior might be suitable.
    method : str, optional
        Which method is to be used. Default is ``ekf0`` which is the
        method proposed by Schober et al.. The available
        options are

        ================================================  ==============
         Extended Kalman filtering/smoothing (0th order)  ``'ekf0'``,
                                                          ``'eks0'``
         Extended Kalman filtering/smoothing (1st order)  ``'ekf1'``,
                                                          ``'eks1'``
         Unscented Kalman filtering/smoothing             ``'ukf'``,
                                                          ``'uks'``
        ================================================  ==============

        First order extended Kalman filtering and smoothing methods
        require Jacobians of the RHS-vector field of the IVP. The
        uncertainty estimates as returned by EKF1/S1 and UKF/S appear to
        be more reliable than those of EKF0/S0. The latter is more
        stable when it comes to very small steps.

    firststep : float, optional
        First suggested step :math:`h_0` for adaptive step size scheme.
        Default is None which lets the solver start with the suggestion
        :math:`h_0 = T - t_0`. For low accuracy it might be more
        efficient to start out with smaller :math:`h_0` so that the
        first acceptance occurs earlier.

    precond_step : float, optional
        Expected average step size, used for preconditioning.
        See :class:`ODEPrior` for details.
        Default is ``precond_step=1.0`` which amounts to no
        preconditioning. If constant step size, precond_step is
        overwritten with the actual step size to provide optimal
        preconditioning.

    Returns
    -------
    solution : ODESolution
        Solution of the ODE problem.

        Contains fields:

        t : :obj:`np.ndarray`, shape=(N,)
            Mesh used by the solver to compute the solution.
            It includes the initial time :math:`t_0` but not necessarily the
            final time :math:`T`.
        y : :obj:`list` of :obj:`RandomVariable`, length=N
            Discrete-time solution at times :math:`t_1, ..., t_N`,
            as a list of random variables.
            The means and covariances can be accessed with ``solution.y.mean``
            and ``solution.y.cov``.

    See Also
    --------
    GaussianIVPFilter : Solve IVPs with Gaussian filtering and smoothing
    ODESolution : Solution of ODE problems

    References
    ----------
    .. [1] Schober, M., Särkkä, S. and Hennig, P..
        A probabilistic model for the numerical solution of initial
        value problems.
        Statistics and Computing, 2019.
    .. [2] Kersting, H., Sullivan, T.J., and Hennig, P..
        Convergence rates of Gaussian ODE filters.
        2019.
    .. [3] Tronarp, F., Kersting, H., Särkkä, S., and Hennig, P..
        Probabilistic solutions to ordinary differential equations as
        non-linear Bayesian filtering: a new perspective.
        Statistics and Computing, 2019.
    .. [4] Tronarp, F., Särkkä, S., and Hennig, P..
        Bayesian ODE solvers: the maximum a posteriori estimate.
        2019.


    Examples
    --------
    >>> from probnum.diffeq import logistic, probsolve_ivp
    >>> from probnum import random_variables as rvs
    >>> initrv = rvs.Dirac(0.15)
    >>> ivp = logistic(timespan=[0., 1.5], initrv=initrv, params=(4, 1))
    >>> solution = probsolve_ivp(ivp, method="ekf0", step=0.1)
    >>> print(solution.y.mean())
    [[0.15      ]
     [0.2076198 ]
     [0.27932997]
     [0.3649165 ]
     [0.46054129]
     [0.55945475]
     [0.65374523]
     [0.73686744]
     [0.8053776 ]
     [0.85895587]
     [0.89928283]
     [0.92882899]
     [0.95007559]
     [0.96515825]
     [0.97577054]
     [0.9831919 ]]

    >>> initrv = rvs.Dirac(0.15)
    >>> ivp = logistic(timespan=[0., 1.5], initrv=initrv, params=(4, 1))
    >>> solution = probsolve_ivp(ivp, method="eks1", which_prior="ioup3", step=0.1)
    >>> print(solution.y.mean())
    [[0.15      ]
     [0.20795795]
     [0.28228416]
     [0.36926   ]
     [0.46646494]
     [0.5658788 ]
     [0.66048505]
     [0.74369892]
     [0.81237394]
     [0.86592791]
     [0.90598293]
     [0.9349573 ]
     [0.95544749]
     [0.96968754]
     [0.97947631]
     [0.98614541]]
    """
    gfilt, firststep, stprl = _create_solver_inputs(
        ivp, method, which_prior, tol, step, firststep, precond_step, **kwargs
    )
    with_smoothing = method[-2] == "s"
    solver = GaussianIVPFilter(ivp, gfilt, with_smoothing=with_smoothing)
    solution = solver.solve(firststep=firststep, steprule=stprl, **kwargs)
    return solution


def _create_solver_inputs(
    ivp, method, which_prior, tol, step, firststep, precond_step, **kwargs
):
    """Create the solver object that is used."""
    _check_step_tol(step, tol)
    _check_method(method)
    if precond_step is None:
        precond_step = step or 1.0
    _prior = _string2prior(ivp, which_prior, precond_step, **kwargs)
    if tol is not None:
        stprl = steprule.AdaptiveSteps(tol, _prior.ordint + 1, **kwargs)
        if firststep is None:
            firststep = ivp.tmax - ivp.t0
    else:
        stprl = steprule.ConstantSteps(step)
        firststep = step
    gfilt = _string2filter(ivp, _prior, method, **kwargs)
    return gfilt, firststep, stprl


def _check_step_tol(step, tol):

    both_none = tol is None and step is None
    both_not_none = tol is not None and step is not None
    if both_none or both_not_none:
        errormsg = "Please specify either a tolerance or a step size."
        raise ValueError(errormsg)


def _check_method(method):

    if method not in ["ekf0", "ekf1", "ukf", "eks0", "eks1", "uks"]:
        raise ValueError("Method not supported.")


def _string2prior(ivp, which_prior, precond_step, **kwargs):

    ibm_family = ["ibm1", "ibm2", "ibm3", "ibm4"]
    ioup_family = ["ioup1", "ioup2", "ioup3", "ioup4"]
    matern_family = ["matern32", "matern52", "matern72", "matern92"]
    if which_prior in ibm_family:
        return _string2ibm(ivp, which_prior, precond_step, **kwargs)
    elif which_prior in ioup_family:
        return _string2ioup(ivp, which_prior, precond_step, **kwargs)
    elif which_prior in matern_family:
        return _string2matern(ivp, which_prior, precond_step, **kwargs)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2ibm(ivp, which_prior, precond_step, **kwargs):

    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if which_prior == "ibm1":
        return prior.IBM(1, ivp.ndim, diffconst, precond_step)
    elif which_prior == "ibm2":
        return prior.IBM(2, ivp.ndim, diffconst, precond_step)
    elif which_prior == "ibm3":
        return prior.IBM(3, ivp.ndim, diffconst, precond_step)
    elif which_prior == "ibm4":
        return prior.IBM(4, ivp.ndim, diffconst, precond_step)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2ioup(ivp, which_prior, precond_step, **kwargs):

    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if "driftspeed" in kwargs.keys():
        driftspeed = kwargs["driftspeed"]
    else:
        driftspeed = 1.0
    if which_prior == "ioup1":
        return prior.IOUP(1, ivp.ndim, driftspeed, diffconst, precond_step)
    elif which_prior == "ioup2":
        return prior.IOUP(2, ivp.ndim, driftspeed, diffconst, precond_step)
    elif which_prior == "ioup3":
        return prior.IOUP(3, ivp.ndim, driftspeed, diffconst, precond_step)
    elif which_prior == "ioup4":
        return prior.IOUP(4, ivp.ndim, driftspeed, diffconst, precond_step)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2matern(ivp, which_prior, precond_step, **kwargs):

    if "diffconst" in kwargs.keys():
        diffconst = kwargs["diffconst"]
    else:
        diffconst = 1.0
    if "lengthscale" in kwargs.keys():
        lengthscale = kwargs["lengthscale"]
    else:
        lengthscale = 1.0
    if which_prior == "matern32":
        return prior.Matern(1, ivp.ndim, lengthscale, diffconst, precond_step)
    elif which_prior == "matern52":
        return prior.Matern(2, ivp.ndim, lengthscale, diffconst, precond_step)
    elif which_prior == "matern72":
        return prior.Matern(3, ivp.ndim, lengthscale, diffconst, precond_step)
    elif which_prior == "matern92":
        return prior.Matern(4, ivp.ndim, lengthscale, diffconst, precond_step)
    else:
        raise RuntimeError("It should have been impossible to reach this point.")


def _string2filter(_ivp, _prior, _method, **kwargs):

    if "evlvar" in kwargs.keys():
        evlvar = kwargs["evlvar"]
    else:
        evlvar = 0.0
    if _method in ("ekf0", "eks0"):
        return ivp2filter.ivp2ekf0(_ivp, _prior, evlvar)
    elif _method in ("ekf1", "eks1"):
        return ivp2filter.ivp2ekf1(_ivp, _prior, evlvar)
    elif _method in ("ukf", "uks"):
        return ivp2filter.ivp2ukf(_ivp, _prior, evlvar)
    else:
        raise ValueError("Type of filter not supported.")



"""
Bayesian filtering and smoothing.
"""

from .statespace import *
from .gaussfiltsmooth import *
from .bayesfiltsmooth import *
from .filtsmoothposterior import FiltSmoothPosterior

# Public classes and functions. Order is reflected in documentation.
__all__ = [
    "GaussFiltSmooth",
    "Kalman",
    "ExtendedKalman",
    "UnscentedKalman",
    "UnscentedTransform",
    "ContinuousModel",
    "LinearSDEModel",
    "LTISDEModel",
    "DiscreteModel",
    "DiscreteGaussianModel",
    "DiscreteGaussianLinearModel",
    "DiscreteGaussianLTIModel",
    "FiltSmoothPosterior",
    "KalmanPosterior",
    "generate_cd",
    "generate_dd",
]

# Set correct module paths (for superclasses).
# Corrects links and module paths in documentation.
ContinuousModel.__module__ = "probnum.filtsmooth"
LinearSDEModel.__module__ = "probnum.filtsmooth"
DiscreteModel.__module__ = "probnum.filtsmooth"
DiscreteGaussianModel.__module__ = "probnum.filtsmooth"
DiscreteGaussianLinearModel.__module__ = "probnum.filtsmooth"
GaussFiltSmooth.__module__ = "probnum.filtsmooth"
KalmanPosterior.__module__ = "probnum.filtsmooth"



import numbers
from typing import Iterable, Tuple, Union

import numpy as np

########################################################################################
# API Types
########################################################################################

ShapeType = Tuple[int, ...]

RandomStateType = Union[np.random.RandomState, np.random.Generator]
"""Type of a random number generator."""

########################################################################################
# Argument Types
########################################################################################

IntArgType = Union[int, numbers.Integral, np.integer]
FloatArgType = Union[float, numbers.Real, np.floating]

ShapeArgType = Union[IntArgType, Iterable[IntArgType]]
"""Type of a public API argument for supplying a shape. Values of this type should
always be converted into :class:`ShapeType` using the function
:func:`probnum.utils.as_shape` before further internal processing."""

DTypeArgType = Union[np.dtype, str]
"""Type of a public API argument for supplying a dtype. Values of this type should
always be converted into :class:`np.dtype` using the function
:func:`np.dtype` before further internal processing."""

ScalarArgType = Union[int, float, complex, numbers.Number, np.float_]
"""Type of a public API argument for supplying a scalar value. Values of this type
should always be converted into :class:`np.generic` using the function
:func:`probnum.utils.as_scalar` before further internal processing."""

ArrayLikeGetitemArgType = Union[
    int,
    slice,
    np.ndarray,
    np.newaxis,
    None,
    type(Ellipsis),
    Tuple[Union[int, slice, np.ndarray, np.newaxis, None, type(Ellipsis)], ...],
]

RandomStateArgType = Union[None, int, np.random.RandomState, np.random.Generator]
"""Type of a public API argument for supplying a random number generator. Values of this
type should always be converted into :class:`RandomStateType` using the function
:func:`probnum.utils.as_random_state` before further internal processing."""



"""
Continuous-Time priors for ODE solvers.

Currently, they are only relevant in the context of ODEs.
If needed in a more general setting, it is easy to move
them to statespace module (->thoughts?)

Matern will be easy to implement, just reuse the template
provided by IOUP and change parameters
"""
import warnings
import numpy as np
from scipy.special import binom  # for Matern
from scipy.special import factorial  # vectorised factorial for IBM-Q(h)

from probnum.filtsmooth.statespace.continuous import LTISDEModel
from probnum.random_variables import Normal


class ODEPrior(LTISDEModel):
    """
    Prior dynamic model for ODE filtering and smoothing.

    An ODE prior is an continuous LTI state space model with attributes:
        * order of integration :math:`q`
        * spatial dimension of the underlying ODE
        * projection to :math:`X_i(t)`
          (the :math:`(i-1)`-th derivative estimate)
        * A preconditioner :math:`P` (see below)

    Instead of the LTI SDE

    .. math:: d X(t) = [F X(t) + u] dt + L dB(t)

    the prior for the ODE Dynamics is given by

    .. math:: dX(t) = P F P^{-1} X(t) dt + P L dB(t)

    where :math:`P` is a preconditioner matrix ensuring stability
    of the iterations. Note that ODE priors do not have a drift term.
    By default, we choose :math:`P` to be the
    matrix that maps to filtering iteration to the Nordsieck vector,

    .. math:: P = \\text{diag }(h^{-q}, h^{-q+1}, ..., 1).

    Here, :math:`h` is some expected average step size. Note that we
    ignored the factorials in this matrix. Our setting makes it easy to
    recover "no preconditioning" by choosing :math:`h=1`.

    - If no expected step size is available we choose :math:`h=1.0`.
      This recovers :math:`P=I_{d(q+1)}`, hence no preconditioning.

    - For fixed step size algorithms this quantity :math:`h` is easy
      to choose

    - For adaptive steps it is a bit more involved.

    Since it doesn't have to be exact, any more or less appropriate
    choice will do well. The main effect of this preconditioning is that
    the predictive covariances inside each filter iteration are
    well-conditioned: for IBM(:math:`q`) priors, the condition number
    of the predictive covariances only depends on order of integration
    :math:`q` and not on the step size anymore. Nb: this only holds
    if all required derivatives of the RHS vector field of the ODE are
    specified: None for IBM(1), Jacobian of :math:`f` for IBM(2),
    Hessian of :math:`f` for IBM(3). If this is not the case the
    preconditioner still helps but is not as powerful anymore.

    Without preconditioning they can be numerically singular for small
    steps and higher order methods which especially makes smoothing
    algorithms unstable.

    Another advantage of this preconditioning is that the smallest value
    that appears inside the algorithm is :math:`h^{q}` (with
    preconditioning) instead of :math:`h^{2q+1}` (without preconditioning).

    The matrices :math:`F, u, L` are the usual matrices for
    IBM(:math:`q`), IOUP(:math:`q`) or Matern(:math:`q+1/2`) processes.
    As always, :math:`B(t)` is
    s-dimensional Brownian motion with unit diffusion matrix :math:`Q`.

    """

    def __init__(self, driftmat, dispmat, ordint, spatialdim, precond_step=1.0):
        self.ordint = ordint
        self.spatialdim = spatialdim
        self.precond, self.invprecond = self.precond2nordsieck(precond_step)
        driftmat = self.precond @ driftmat @ self.invprecond
        dispmat = self.precond @ dispmat
        forcevec = np.zeros(len(driftmat))
        diffmat = np.eye(spatialdim)
        super().__init__(driftmat, forcevec, dispmat, diffmat)

    def proj2coord(self, coord):
        """
        Projection matrix to :math:`i`-th coordinates.

        Computes the matrix

        .. math:: H_i = \\left[ I_d \\otimes e_i \\right] P^{-1},

        where :math:`e_i` is the :math:`i`-th unit vector,
        that projects to the :math:`i`-th coordinate of a vector.
        If the ODE is multidimensional, it projects to **each** of the
        :math:`i`-th coordinates of each ODE dimension.

        Parameters
        ----------
        coord : int
            Coordinate index :math:`i` which to project to.
            Expected to be in range :math:`0 \\leq i \\leq q + 1`.

        Returns
        -------
        np.ndarray, shape=(d, d*(q+1))
            Projection matrix :math:`H_i`.
        """
        projvec1d = np.eye(self.ordint + 1)[:, coord]
        projmat1d = projvec1d.reshape((1, self.ordint + 1))
        projmat = np.kron(np.eye(self.spatialdim), projmat1d)
        projmat1d_with_precond = projmat @ self.invprecond
        return projmat1d_with_precond

    def precond2nordsieck(self, step):
        """
        Computes preconditioner inspired by Nordsieck.

        Computes the matrix :math:`P` given by

        .. math:: P = I_d \\otimes diag (1, h, h^2, ..., h^q)

        as well as its inverse :math:`P^{-1}`.

        Parameters
        ----------
        step : float
            Step size :math:`h` used for preconditioning. If :math:`h`
            is so small that :math:`h^q! < 10^{-15}`, it is being
            set to :math:`h = (\\cdot 10^{-15})^{1/q}`.

        Returns
        -------
        precond : np.ndarray, shape=(d(q+1), d(q+1))
            Preconditioner matrix :math:`P`.
        invprecond : np.ndarray, shape=(d(q+1), d(q+1))
            Inverse preconditioner matrix :math:`P^{-1}`.
        """
        smallval = step ** self.ordint
        if smallval < 1e-15:
            warnmsg = (
                "Preconditioner contains values below "
                "machine precision (%.1e)" % smallval
            )
            warnings.warn(message=warnmsg, category=RuntimeWarning)
            step = 1e-15 ** (1 / self.ordint)
        powers = np.arange(start=-self.ordint, stop=1)
        diags = step ** powers
        precond = np.kron(np.eye(self.spatialdim), np.diag(diags))
        invprecond = np.kron(np.eye(self.spatialdim), np.diag(1.0 / diags))
        return precond, invprecond

    @property
    def preconditioner(self):
        """
        Convenience property to return the readily-computed
        preconditioner without having to remember abbreviations.

        Returns
        -------
        np.ndarray, shape=(d(q+1), d(q+1))
            Preconditioner matrix :math:`P`
        """
        return self.precond

    @property
    def inverse_preconditioner(self):
        """
        Convenience property to return the readily-computed
        inverse preconditioner without having to remember abbreviations.

        Returns
        -------
        np.ndarray, shape=(d(q+1), d(q+1))
            Inverse preconditioner matrix :math:`P^{-1}`
        """
        return self.invprecond


class IBM(ODEPrior):
    """
    Integrated Brownian motion of order :math:`q` prior.

    The integrated Brownian motion prior is represented through
    the LTI SDE

    .. math:: dX(t) =  F X(t) dt + L dB(t)

    where for readibility reasons we did not write the preconditioner
    matrix :math:`P`; see :class:`ODEPrior` for explanations.

    - It has driftmatrix :math:`F` given by

      .. math:: F = I_d \\otimes \\tilde F, \\quad
        \\tilde F = \\begin{pmatrix} 0 & I_q \\\\ 0 & 0 \\end{pmatrix}

      where the top left zero-vector has :math:`q` rows and 1 column.

    - It has dispersion matrix :math:`L` given by

      .. math:: L = I_d \\otimes \\tilde L, \\quad
        \\tilde L = \\sigma \\, e_{q+1}

      where :math:`\\sigma` is the diffusion constant, that is,
      :math:`\\sigma^2` is the intensity of each dimension of the
      :math:`d`-dimensional Brownian motion driving the SDE and
      :math:`e_{q+1}=(0, ..., 0, 1)` is the :math:`(q+1)`-st unit vector.

    - The Brownian motion :math:`B=B(t)` driving the SDE has unit
      diffusion :math:`Q = I_d`.

    Parameters
    ----------
    ordint : int
        Order of integration :math:`q`. The higher :math:`q`, the higher
        the order of the ODE filter.
    spatialdim : int
        Spatial dimension :math:`d` of the ordinary differential
        equation that is to be modelled.
    diffconst : float
        Diffusion constant :math:`sigma` of the stochastic process.
    precond_step : float, optional
        Expected step size :math:`h` used in the ODE filter.
        This quantity is used for preconditioning, see :class:`ODEPrior`
        for a clear explanation. Default is :math:`h=1`.
    """

    def __init__(self, ordint, spatialdim, diffconst, precond_step=1.0):
        """
        ordint : this is "q"
        spatialdim : d
        diffconst : sigma
        """
        self.diffconst = diffconst
        driftmat = _driftmat_ibm(ordint, spatialdim)
        dispmat = _dispmat(ordint, spatialdim, diffconst)
        super().__init__(driftmat, dispmat, ordint, spatialdim, precond_step)

    def chapmankolmogorov(self, start, stop, step, randvar, *args, **kwargs):
        """
        Closed form solution to the Chapman-Kolmogorov equations
        for the integrated Brownian motion.

        It is given by

        .. math:: X_{t+h} \\, | \\, X_t \\sim \\mathcal{N}(A(h)X_t, Q(h))

        with matrices :math:`A(h)` and `Q(h)` defined by

        .. math:: [A(h)]_{ij} = \\mathbb{I}_{i\\leq j} \\frac{h^{j-i}}{(j-i)!}


        .. math:: [Q(h)]_{ij} = \\sigma^2 \\frac{h^{2q+1-i-j}}{(2q+1-i-j)!(q-j)!(q-i)!}


        The implementation that is used here is more stable than the matrix-exponential
        implementation in :meth:`super().chapmankolmogorov` which is relevant for
        combinations of large order :math:`q` and small steps :math:`h`.
        In these cases even the preconditioning is subject to numerical
        instability if the transition matrices :math:`A(h)`
        and :math:`Q(h)` are computed with matrix exponentials.

        "step" variable is obsolent here and is ignored.
        """
        mean, covar = randvar.mean, randvar.cov
        ah = self._trans_ibm(start, stop)
        qh = self._transdiff_ibm(start, stop)
        mpred = ah @ mean
        crosscov = covar @ ah.T
        cpred = ah @ crosscov + qh
        return Normal(mpred, cpred), crosscov

    def _trans_ibm(self, start, stop):
        """
        Computes closed form solution for the transition matrix A(h).
        """
        step = stop - start

        # This seems like the faster solution compared to fully vectorising.
        # I suspect it is because np.math.factorial is much faster than
        # scipy.special.factorial
        ah_1d = np.diag(np.ones(self.ordint + 1), 0)
        for i in range(self.ordint):
            offdiagonal = (
                step ** (i + 1) / np.math.factorial(i + 1) * np.ones(self.ordint - i)
            )
            ah_1d += np.diag(offdiagonal, i + 1)

        ah = np.kron(np.eye(self.spatialdim), ah_1d)
        return self.precond @ ah @ self.invprecond

    def _transdiff_ibm(self, start, stop):
        """
        Computes closed form solution for the diffusion matrix Q(h).
        """
        step = stop - start
        indices = np.arange(0, self.ordint + 1)
        col_idcs, row_idcs = np.meshgrid(indices, indices)
        exponent = 2 * self.ordint + 1 - row_idcs - col_idcs
        factorial_rows = factorial(
            self.ordint - row_idcs
        )  # factorial() handles matrices but is slow(ish)
        factorial_cols = factorial(self.ordint - col_idcs)
        qh_1d = (
            self.diffconst ** 2
            * step ** exponent
            / (exponent * factorial_rows * factorial_cols)
        )
        qh = np.kron(np.eye(self.spatialdim), qh_1d)
        return self.precond @ qh @ self.precond.T


def _driftmat_ibm(ordint, spatialdim):
    """
    Returns I_d \\otimes F
    """
    driftmat_1d = np.diag(np.ones(ordint), 1)
    return np.kron(np.eye(spatialdim), driftmat_1d)


class IOUP(ODEPrior):
    """
    IOUP(q) prior:

    F = I_d \\otimes F
    L = I_d \\otimes L = I_d \\otimes (0, ...,  diffconst**2)
    Q = I_d
    """

    def __init__(self, ordint, spatialdim, driftspeed, diffconst, precond_step=1.0):
        """
        ordint : this is "q"
        spatialdim : d
        driftspeed : float > 0; (lambda; note that -lambda ("minus"-lambda)
            is used in the OU equation!!
        diffconst : sigma
        """
        self.driftspeed = driftspeed
        self.diffconst = diffconst
        driftmat = _driftmat_ioup(ordint, spatialdim, self.driftspeed)
        dispvec = _dispmat(ordint, spatialdim, diffconst)
        super().__init__(driftmat, dispvec, ordint, spatialdim, precond_step)


def _driftmat_ioup(ordint, spatialdim, driftspeed):
    """
    Returns I_d \\otimes F
    """
    driftmat = np.diag(np.ones(ordint), 1)
    driftmat[-1, -1] = -driftspeed
    return np.kron(np.eye(spatialdim), driftmat)


class Matern(ODEPrior):
    """
    Matern(q) prior --> Matern process with reg. q+0.5
    and hence, with matrix size q+1

    F = I_d \\otimes F
    L = I_d \\otimes L = I_d \\otimes diffconst*(0, ..., 1)
    Q = I_d
    """

    def __init__(self, ordint, spatialdim, lengthscale, diffconst, precond_step=1.0):
        """
        ordint : this is "q"
        spatialdim : d
        lengthscale : used as 1/lengthscale, remember that!
        diffconst : sigma

        """
        self.lengthscale = lengthscale
        self.diffconst = diffconst
        driftmat = _driftmat_matern(ordint, spatialdim, self.lengthscale)
        dispvec = _dispmat(ordint, spatialdim, diffconst)
        super().__init__(driftmat, dispvec, ordint, spatialdim, precond_step)


def _driftmat_matern(ordint, spatialdim, lengthscale):
    """
    Returns I_d \\otimes F
    """
    driftmat = np.diag(np.ones(ordint), 1)
    nu = ordint + 0.5
    D, lam = ordint + 1, np.sqrt(2 * nu) / lengthscale
    driftmat[-1, :] = np.array([-binom(D, i) * lam ** (D - i) for i in range(D)])
    return np.kron(np.eye(spatialdim), driftmat)


def _dispmat(ordint, spatialdim, diffconst):
    """
    Returns I_D \\otimes L
    diffconst = sigma**2
    """
    dispvec = diffconst * np.eye(ordint + 1)[:, -1]
    return np.kron(np.eye(spatialdim), dispvec).T
